#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "../libraryStable/libraryBasic.h"
#include "../libraryStable/libsift.h"




#include "libburst.h"
#include "poisson_fft.h"



/*
 
 // compression: compression factor. So width1 is converted into width1 / compression
 // Procedure:
 // -
 
 void apply_homography_picture_anti_aliasing(float *input, int width1, int height1, float **H, float bg, int order, float *out, float x0, float y0, int nwidth, int nheight, float compression , int zoom)
 {
 
 
 // Modify homography in order to take into account the zoom
 float zoomIn = compression;
 
 float **modH = allocate_float_matrix(3, 3);
 for (int ii=0; ii < 3; ii++) modH[0][ii] = zoomIn * H[0][ii];
 for (int ii=0; ii < 3; ii++) modH[1][ii] = zoomIn * H[1][ii];
 for (int ii=0; ii < 3; ii++) modH[2][ii] =  H[2][ii];
 
 
 // apply homography without compression
 int znwidth = (int) (zoomIn * (float) nwidth);
 int znheight = (int) (zoomIn * (float) nheight);
 
 float *tmp = new float[znwidth * znheight];
 fpClear(tmp, 0.0f, znwidth*znheight);
 
 apply_homography_picture(input,width1,height1,modH, bg,order, tmp ,x0,y0, znwidth, znheight, 1);	
 
 
 // make now subsampling of factor compression taking into account that there's a zoom parameter
 float sigma = 0.8*sqrt(zoomIn*zoomIn-1.0f);
 
 fiFloatGaussianImageConvolution(tmp, tmp, znwidth, znheight, sigma);
 
 //float subsampling = compression / (float) zoom;
 
 fiImageSample(tmp, out, compression,  znwidth, znheight);
 
 delete[] tmp; // 21 oct. 2010
 desallocate_float_matrix(modH,3,3); // 21 oct. 2010
 
 }
 
 
 
 
 */


int main(int argc, char **argv)
{
	
	
    std::vector <OptStruct *> options;
    OptStruct oV  = {"v", 0, NULL, NULL, "verbose mode"}; options.push_back(&oV);	
    OptStruct oD  = {"d", 0, NULL, NULL, "apply distorsion correction"}; options.push_back(&oD);	
    OptStruct oL  = {"l", 0, NULL, NULL, "apply convex hull restriction"}; options.push_back(&oL);	
    OptStruct oA  = {"a:", 0, NULL, NULL, "0: my Sift, 1: Sift IPOL, 2: Asift IPOL"}; options.push_back(&oA);	
    OptStruct oS  = {"s", 0, NULL, NULL, "use scale fusion"}; options.push_back(&oS);	
    
    
    
    
	std::vector<ParStruct *> pparameters;
	ParStruct pinput = {"input", NULL, "input file"}; pparameters.push_back(&pinput);
	ParStruct pout = {"out", NULL, "output file"}; pparameters.push_back(&pout);
	
	
	if (!parsecmdline("iip_burst_fusion_by_scale","burst fusion of single resolution images", argc, argv, options, pparameters))
		return 0;
	
	
    
	
	//--//  Parameters  // -- //
    int iVerbose  = oV.flag; 
	
    
	int iDistCorrectionFlag     = oD.flag;
	int iAsiftFlag     = atoi(oA.value);
    int iScalesFlag = oS.flag;
    int iDelaunayFlag = oL.flag;
    
    
	
	//--//   Open File  // -- //
    std::ifstream f(pinput.value);
	if (!f.good()) 
	{
		std::cout << "input file " << pinput.value << "not found or impossible to open\n"; return 0;
	}
	
	
	int inBursts;
	f >> inBursts;
	std::cout << "... Total number of images: " << inBursts << std::endl  << std::endl ;
	
	
	
	
	//--//   Read Images  // -- //
	
	libIIPStable::cflimage *cfBursts = new libIIPStable::cflimage[inBursts];
	for(int iiIndex = 0; iiIndex < inBursts; iiIndex++)
	{
		char *strFilename = new char[256];
		f >> strFilename;											
		cfBursts[iiIndex].load(strFilename);
	}
	
	
    
    
    //--// Memory for registered images //--//
    libIIPStable::cflimage *cfRBursts = new libIIPStable::cflimage[inBursts];
    for (int kk=0; kk < inBursts; kk++) cfRBursts[kk] = cfBursts[kk];
    
    
    
    if (!iScalesFlag)
    {
        
        if (iVerbose)
        {
            std::cout << "... Performing fusion with single scale" << std::endl;
        }
        
        
        //! Index of correctly registered images
        int  iCorrectlyRegistered;
        int  *ipCorrectlyRegistered = new int[inBursts];
        
        //! Scale processing for single burst if activated (no activated)
        int iCompressionFlag=0;
        float lScale=-1.0f, uScale=-1.0f;
        
        register_burst_with_initial_image(cfBursts, cfRBursts, 0, inBursts, 0, 0, iAsiftFlag, iVerbose, iCorrectlyRegistered, ipCorrectlyRegistered,iCompressionFlag, lScale, uScale );
        
        
        
    	
        //--//  Re - Register each image with initial one with distorsion // -- //
        
        if (iVerbose) std::cout <<  std::endl << "... reregistering bursts" << std::endl << std::endl;
		
        for (int kk=0; kk < inBursts; kk++) cfBursts[kk] = cfRBursts[kk];
        register_burst_with_initial_image(cfBursts, cfRBursts, 0, inBursts, iDistCorrectionFlag, iDelaunayFlag, 0, iVerbose, iCorrectlyRegistered, ipCorrectlyRegistered,iCompressionFlag, lScale, uScale);
        
        
        
        
        //--//  Histogram specification  // -- //
        
        if (iVerbose)	std::cout <<  std::endl << "... specifying histograms for registered bursts" << std::endl << std::endl ;
        for (int ii=1; ii < inBursts; ii++)
            for (int iC=0; iC < cfBursts[0].c(); iC++)
            {
				
				libIIPStable::fk_histogram_specification(cfRBursts[0].v(iC), cfRBursts[ii].v(iC), cfRBursts[ii].v(iC),  cfRBursts[0].w(),  cfRBursts[0].h(), cfRBursts[ii].w(), cfRBursts[ii].h());
				
            }
        
        
        
        
        //--//  Save re - registered Bursts  // -- //
        if (iVerbose)
        {
            for (int ii=0; ii < inBursts; ii++)
            {
                std::string filename(pout.value); filename += "_registered_bursts_";
                char numero[128]; sprintf(numero, "%d", ii);
                std::string snumero(numero); filename += snumero; filename += ".png";
                
                cfRBursts[ii].save(filename.c_str());
            }
        }
        
        
        
        
        //--//  Combine registered bursts  // -- //
        
        if (iVerbose)	std::cout << std::endl << "... combining bursts " << std::endl;
        
        libIIPStable::cflimage cfCombined;
        combine_registered_bursts_min(cfRBursts, inBursts, cfCombined);
        std::string filename_min(pout.value); filename_min += "_min.png";
        cfCombined.save(filename_min.c_str());
        
        combine_registered_bursts_medianRGB(cfRBursts, inBursts, cfCombined);
        std::string filename_median(pout.value); filename_median += "_median.png";
        cfCombined.save(filename_median.c_str());
        
        combine_registered_bursts_medianGrad(cfRBursts, inBursts, cfCombined);
        std::string filename_grad(pout.value); filename_grad += "_grad.png";
        cfCombined.save(filename_grad.c_str());
        
        
        
        
        
    } else
    {
        
        
        if (iVerbose)
        {
            std::cout << "... Performing fusion with multiple scales" << std::endl;
        }
        
        
        //--//  Perform fusion by scales //--//
        int iScaleNum = 4;
        float *lScale = new float[iScaleNum];
        float *uScale = new float[iScaleNum];
        
        lScale[0] = 0.8f;
        uScale[0] = 1.5f;
        
        lScale[1] = 0.3f;
        uScale[1] = 0.8f;
        
        lScale[2] = 0.1f;
        uScale[2] = 0.3f;

        
        lScale[3] = 0.0f;
        uScale[3] = 0.1f;

        
        
        
        int iProcessedImages = 1;   
        int iScale = 0;
        while ( inBursts > 1 /*iProcessedImages <= inBursts*/ && iScale < iScaleNum)
        {
            
            
            
            //--//  Register each image with initial one // -- //
            
            if (iVerbose) std::cout  <<  std::endl <<  std::endl <<  std::endl << "\t... registering bursts scale "   <<  iScale <<  " = (" << lScale[iScale] << " , " <<  uScale[iScale] <<  ")" << std::endl << std::endl;
            
            
            //! Index of correctly registered frames
            int  iCorrectlyRegistered;
            int  *ipCorrectlyRegistered = new int[inBursts];
            
            
            int iCompressionFlag = 1;
            register_burst_with_initial_image(cfBursts, cfRBursts, 0, inBursts, 0, 0, iAsiftFlag, iVerbose, iCorrectlyRegistered, ipCorrectlyRegistered, iCompressionFlag, lScale[iScale], uScale[iScale]);
            
            
            //! Print index of correctly registered
            if (iVerbose)
            {    
                std::cout << "\t... correctly registered: " << iCorrectlyRegistered << " -->  " ;
                for (int ii=0; ii < iCorrectlyRegistered; ii++) std::cout << ipCorrectlyRegistered[ii] << ", " ;
                std::cout << std::endl;
            }
            
            
            
            //! Memory for registered images
            if (iCorrectlyRegistered > 1)
            {
                
                
                
                libIIPStable::cflimage *cflSelected = new  libIIPStable::cflimage[iCorrectlyRegistered];
                libIIPStable::cflimage *cflRSelected = new  libIIPStable::cflimage[iCorrectlyRegistered];
                
                
                //! Place registered into a new vector and keep non registered in CfBursts
                int *iAuxIndex = new int[inBursts];
                for (int ii=0; ii < inBursts; ii++)  iAuxIndex[ii] = 0;
                for (int ii=0; ii < iCorrectlyRegistered; ii++)  iAuxIndex[ (int) ipCorrectlyRegistered[ii] ] = 1;
                
                
                int iCounterNR = 1;         // We count the reference as not registered for updating cfBursts even if iAuxIndex[0]=1;
                // we'll place the updated in position 0
                
                
                int iCounterR = 1;          // The reference is also used in order to perform the fusion
                cflSelected[0] = cfBursts[0];          
                for (int ii = 1; ii < inBursts; ii++)
                {
                    if (iAuxIndex[ii] == 0) { cfBursts[iCounterNR] = cfRBursts[ii]; iCounterNR++;}
                    
                    if (iAuxIndex[ii] == 1) { cflSelected[iCounterR] = cfRBursts[ii]; iCounterR++;}
                }
                
                
                
                
                
                
                
                //--//  Re - Register each image with initial one with distorsion // -- //
                if (iVerbose) std::cout <<  std::endl << "\t... reregistering bursts" << std::endl << std::endl;
                
                iCompressionFlag = 0;
                register_burst_with_initial_image(cflSelected, cflRSelected, 0, iCounterR, iDistCorrectionFlag, iDelaunayFlag, 0, iVerbose, iCorrectlyRegistered, ipCorrectlyRegistered, iCompressionFlag, lScale[iScale], uScale[iScale]);
                
                
                
                
                //--//  Save re-registered Bursts  // -- //
                if (iVerbose)
                {
                    
                    for (int ii=0; ii < iCounterR; ii++)
                    {
                        
                        std::string filename(pout.value); filename += "_reregistered_bursts_";
                        char numeros[128]; sprintf(numeros, "%d", iScale);
                        std::string snumeros(numeros); filename += snumeros; filename += "_";
                        
                        
                        char numero[128]; sprintf(numero, "%d", ii);
                        std::string snumero(numero); filename += snumero; 
                        
                        cflRSelected[ii].save(filename.c_str());
                    }
                    
                }
                
                
                
                
                
                //--//  Histogram specification  // -- //
                
                if (iVerbose)	std::cout <<  std::endl << "\t... specifying histograms for registered bursts" << std::endl << std::endl ;
                for (int ii=1; ii < iCounterR; ii++)
                    for (int iC=0; iC < cfBursts[0].c(); iC++)
                    {
                        
                        libIIPStable::histogram_specification_painting(cflRSelected[0].v(iC), cflRSelected[ii].v(iC), cflRSelected[ii].v(iC),  cflRSelected[0].w(),  cflRSelected[0].h(), cflRSelected[ii].w(), cflRSelected[ii].h());
                        
                    }
                
                
                
                
                
                
                
                //--//  Combine re-registered bursts  // -- //
                
                if (iVerbose)	std::cout << std::endl << "\t... combining bursts " << std::endl;
                
                
                if (iCounterR > 1 && iScale==0)
                {
                    libIIPStable::cflimage cfCombined;
                    combine_registered_bursts_medianGrad(cflRSelected, iCounterR, cfCombined);
                    cfBursts[0] = cfCombined;
                    
                } else if (iCounterR > 1)
                {    
                    libIIPStable::cflimage cfCombined;
                    combine_registered_bursts_medianGrad_dif_res(cflRSelected, iCounterR, cfCombined);
                    cfBursts[0] = cfCombined;
                    
                }
                else 
                    cfBursts[0] = cfRBursts[0];
                
                
                
                
                //! Save intermediate image
                {
                    std::string filename(pout.value); filename += "_intermediate_";
                    char numeros[128]; sprintf(numeros, "%d", iScale);
                    std::string snumeros(numeros); filename += snumeros; 
                    
                    cfBursts[0].save(filename.c_str());
                }
                
                
                
                
                
                inBursts = iCounterNR;
                iProcessedImages += (iCounterR - 1);
                delete[] cflSelected;
                delete[] cflRSelected;
                
            }
            
            
            //! Update scale
            iScale++;
            
            
            //! delete memory
            delete[] ipCorrectlyRegistered;
            
        }
        
        
        
        
    }
    
    
}





/*
 //--//  Save re - registered Bursts  // -- //
 if (iVerbose)
 {
 for (int ii=0; ii < inBursts; ii++)
 {
 std::string filename(pout.value); filename += "_registered_bursts_";
 char numero[128]; sprintf(numero, "%d", ii);
 std::string snumero(numero); filename += snumero; //filename += ".png";
 
 cfRBursts[ii].save(filename.c_str());
 }
 }
 
 
 
 
 
 
 //--//  Combine registered bursts  // -- //
 
 if (iVerbose)	std::cout << std::endl << "... combining bursts " << std::endl;
 
 libIIPStable::cflimage cfCombined;
 
 
 combine_registered_bursts_medianRGB_dif_res(cfRBursts, inBursts, cfCombined);
 std::string filename_median(pout.value); filename_median += "_median.png";
 cfCombined.save(filename_median.c_str());
 
 
 
 combine_registered_bursts_medianGrad_dif_res(cfRBursts, inBursts, cfCombined);
 std::string filename_grad(pout.value); filename_grad += "_grad";
 cfCombined.save(filename_grad.c_str());
 
 
 }
 */
