#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "../libraryStable/libraryBasic.h"
#include "../libraryStable/libsift.h"




#include "libburst.h"
#include "poisson_fft.h"






int main(int argc, char **argv)
{
	
	
    std::vector <OptStruct *> options;
	OptStruct oS  = {"s", 0, NULL, NULL, "force single burst"}; options.push_back(&oS);	
    //	OptStruct oP  = {"p", 0, NULL, NULL, "pregroup with subsampled image"}; options.push_back(&oP);	
    OptStruct oA  = {"a", 0, NULL, NULL, "adaptive average inside each burst"}; options.push_back(&oA);	
    OptStruct oV  = {"v", 0, NULL, NULL, "verbose"}; options.push_back(&oV);	
	OptStruct oD  = {"d", 0, NULL, NULL, "apply distorsion correction"}; options.push_back(&oD);	
	OptStruct oC  = {"c:", 0, "2", NULL, "subsampling for grouping"}; options.push_back(&oC);	
	
	
    std::vector<ParStruct *> pparameters;
	ParStruct pinput = {"input", NULL, "input file"}; pparameters.push_back(&pinput);
	ParStruct pout = {"out", NULL, "output file"}; pparameters.push_back(&pout);
	
	
	if (!parsecmdline("iip_burst_group_denoise","burst fusion of single resolution images", argc, argv, options, pparameters))
		return 0;
	
	
    
	
	//--//  Parameters  // -- //
    
    int iSingleBurstFlag        = oS.flag;
    int iAdaptiveBurst          = oA.flag;
    int iVerbose				= oV.flag;
    int iDistCorrectionFlag     = oD.flag;
    //    int iPrecomputGroupsFlag    = oP.flag;
	int iSubsampling = atoi(oC.value);
	
	
	//--//   Open File  // -- //
    
    std::ifstream f(pinput.value);
	if (!f.good()) 
	{
		std::cout << "input file " << pinput.value << "not found or impossible to open\n"; return 0;
	}
	
	
	int inImages;
	f >> inImages;
	std::cout << "... Total number of images: " << inImages << std::endl  << std::endl ;
	
	
	
	
	
	//--//   Read Images  // -- //
	
	libIIPStable::cflimage *idata = new libIIPStable::cflimage[inImages];
	for(int iiIndex = 0; iiIndex < inImages; iiIndex++)
	{
		char *strFilename = new char[256];
		f >> strFilename;											
		idata[iiIndex].load(strFilename);
	}
	
	
	
	
	//--//   Group Bursts and register images  // -- //
	
	int inBursts;
    libIIPStable::cflimage *cfBursts;
    
    
	//! Registered data
	libIIPStable::cflimage *rdata = new libIIPStable::cflimage[inImages];
	
	
	//! Index of grouped bursts
	int *fiIndex = new int[inImages]; 
	int *fiNumber = new int[inImages];
	
    
	if(iSingleBurstFlag)
	{
        
        //! Index of correctly registered frames
        int  iCorrectlyRegistered;
        int  *ipCorrectlyRegistered = new int[inImages];
        
		inBursts = 1;
		fiIndex[0] = 0;
        
		//! Scale processing for single burst if activated (no activated)
        int iCompressionFlag=0;
        float lScale=0, uScale=0;
        
		register_burst_with_initial_image(idata, rdata, 0, inImages, iDistCorrectionFlag, 0,  0, iVerbose, iCorrectlyRegistered, ipCorrectlyRegistered,iCompressionFlag, lScale, uScale);
        
	} else
    {
        
        build_groups_of_images(idata, inImages, inBursts, fiIndex, fiNumber, iVerbose, iSubsampling);
        
        for (int ii=0; ii < inBursts; ii++)
        {
            
            
            //! Index of correctly registered frames
            int  iCorrectlyRegistered;
            int  *ipCorrectlyRegistered = new int[fiNumber[ii]];
            
            //! Scale processing for single burst if activated (no activated)
            int iCompressionFlag=0;
            float lScale=0, uScale=0;
            
            register_burst_with_initial_image(idata, rdata, fiIndex[ii], fiNumber[ii], iDistCorrectionFlag,0, 0, iVerbose, iCorrectlyRegistered, ipCorrectlyRegistered, iCompressionFlag, lScale, uScale);            
            
            delete[] ipCorrectlyRegistered;
            
        }
        
        
        
	}
	
	
	std::cout << std::endl << "\t... " << inBursts << " bursts build: \t" ;
    for (int ii=0; ii < inBursts; ii++)  std::cout << fiIndex[ii] << "\t";
    std::cout << std::endl;
	
	
	
	
	//--//  Histogram specification if adaptive average chosen  // -- //
    
    if (iVerbose) std::cout <<  std::endl << "\t... specifying histograms for each burst " << std::endl ;
    
    
#ifdef _OPENMP
#pragma omp parallel shared(rdata, fiIndex)
#endif
    {
        
        
#ifdef _OPENMP
#pragma omp for schedule(dynamic) nowait
#endif
    for (int ii=0; ii < inBursts; ii++)
        
        for (int jj = fiIndex[ii]+1; (ii != inBursts-1 && jj < fiIndex[ii+1]) || (ii == inBursts-1 && jj < inImages); jj++) 
        {
            
            for (int iC=0; iC < rdata[fiIndex[ii]].c(); iC++)
                libIIPStable::fk_histogram_specification(rdata[fiIndex[ii]].v(iC), rdata[jj].v(iC), rdata[jj].v(iC),  rdata[fiIndex[ii]].w(),  rdata[fiIndex[ii]].h(), rdata[jj].w(), rdata[jj].h());
            
        }
    
	
    }
    
    
    
    
	
	//--//   Save registered images // -- //
	if (iVerbose)
	{		
		
		for (int ii=0; ii < inImages; ii++)
		{    
            std::string filename(pout.value); filename += "_registered_"; char numero[128];
			sprintf(numero, "%d", ii);	std::string snumero(numero); filename += snumero; filename += ".png";                
			rdata[ii].save(filename.c_str()); 
		}
	}
	
	
	
	
	
	
	//--//   Average Bursts  // -- //
	cfBursts = new libIIPStable::cflimage[inBursts];
	
	if (iVerbose && iAdaptiveBurst) std::cout <<  std::endl << "\t... averaging bursts taking into account blur " << std::endl;
	else if (iVerbose) std::cout <<   std::endl << "\t... averaging bursts " << std::endl;
    
	
	average_already_registered_burst(rdata, cfBursts, inBursts, fiIndex, inImages, iAdaptiveBurst, iVerbose);
    
    
	
	
	//--//   Save  Bursts  // -- // 
	for (int ii=0; ii < inBursts; ii++)
	{
        std::string filename(pout.value) ; filename += "_burst_";
		char numero[128]; sprintf(numero, "%d", ii);
		std::string snumero(numero); filename += snumero; //filename += ".png";
		
		cfBursts[ii].save(filename.c_str());
	}
	
	
}
