#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "../libraryStable/libraryBasic.h"
#include "../libraryStable/libsift.h"




#include "libburst.h"
#include "poisson_fft.h"




int main(int argc, char **argv)
{
	
	
    std::vector <OptStruct *> options;
	OptStruct oS  = {"s", 0, NULL, NULL, "force single burst"}; options.push_back(&oS);	
    //	OptStruct oP  = {"p", 0, NULL, NULL, "pregroup with subsampled image"}; options.push_back(&oP);	
    OptStruct oA  = {"a", 0, NULL, NULL, "adaptive average inside each burst"}; options.push_back(&oA);	
    OptStruct oV  = {"v", 0, NULL, NULL, "verbose"}; options.push_back(&oV);	
	OptStruct oD  = {"d", 0, NULL, NULL, "apply distorsion correction"}; options.push_back(&oD);	
	OptStruct oC  = {"c:", 0, "2", NULL, "subsampling for grouping"}; options.push_back(&oC);	
	
	
    std::vector<ParStruct *> pparameters;
	ParStruct pinput = {"input", NULL, "input file"}; pparameters.push_back(&pinput);
	ParStruct pout = {"out", NULL, "output file"}; pparameters.push_back(&pout);
	
	
	if (!parsecmdline("iip_burst_group_denoise","burst fusion of single resolution images", argc, argv, options, pparameters))
		return 0;
	
	
    
	
	//--//  Parameters  // -- //
    
    int iSingleBurstFlag        = oS.flag;
    int iAdaptiveBurst          = oA.flag;
    int iVerbose				= oV.flag;
    int iDistCorrectionFlag     = oD.flag;
    //    int iPrecomputGroupsFlag    = oP.flag;
	int iSubsampling = atoi(oC.value);
	
	
	//--//   Open File  // -- //
    
    std::ifstream f(pinput.value);
	if (!f.good()) 
	{
		std::cout << "input file " << pinput.value << "not found or impossible to open\n"; return 0;
	}
	
	
	int inImages;
	f >> inImages;
	std::cout << "... Total number of images: " << inImages << std::endl  << std::endl ;
    
    
    //--//   Read Names of Images  // -- //
	
	char **idata = new char*[inImages];
	for(int iiIndex = 0; iiIndex < inImages; iiIndex++)
	{
		idata[iiIndex] = new char[256];
		f >> idata[iiIndex];	
	}
    
    
    
    //! Index of grouped bursts
	int *fiIndex = new int[inImages]; 
	int *fiNumber = new int[inImages];
    
    
    int inBursts;
    //libIIPStable::cflimage *cfBursts;
    
    
    
	if(iSingleBurstFlag)
	{
        
        
    } else
    {
        
        
        
        libIIPStable::build_groups_of_images_lm(idata, inImages, inBursts, fiIndex, fiNumber, iVerbose, iSubsampling);
        
        std::cout << std::endl << "\t... " << inBursts << " bursts build: \t" ;
        for (int ii=0; ii < inBursts; ii++)  std::cout << fiIndex[ii] << "\t";
        std::cout << std::endl;
        
        
        
        //cfBursts = new libIIPStable::cflimage[inBursts];
        libIIPStable::cflimage cfBurst;
        
        for (int ii=0; ii < inBursts; ii++)
        {
            
            
            //! Index of correctly registered frames
            int  iCorrectlyRegistered;
            int  *ipCorrectlyRegistered = new int[fiNumber[ii]];
            
            
            //! Fill buffer of frames to be registered in current group
            libIIPStable::cflimage *inImData = new libIIPStable::cflimage[fiNumber[ii]]; 
            libIIPStable::cflimage *rImData = new libIIPStable::cflimage[fiNumber[ii]]; 
            
            for (int iik = 0; iik < fiNumber[ii]; iik++) 
            {
                inImData[iik].load(idata[fiIndex[ii] + iik]);
                rImData[iik] = inImData[iik];
            }
            
            
            //! Scale processing for single burst if activated (no activated)
            int iCompressionFlag=0;
            float lScale=0, uScale=0;
            
            // rImData created only for images well registered
            register_burst_with_initial_image(inImData, rImData, 0, fiNumber[ii], iDistCorrectionFlag,0, 0, iVerbose, iCorrectlyRegistered, ipCorrectlyRegistered, iCompressionFlag, lScale, uScale);            
            
            delete[] ipCorrectlyRegistered;
            
            
            //! Histogram equalization for current burst
#ifdef _OPENMP
#pragma omp parallel shared(rImData, fiNumber)
#endif
            {
                
                
#ifdef _OPENMP
#pragma omp for schedule(dynamic) nowait 
#endif
                for (int jj = 1; jj < fiNumber[ii]; jj++) 
                {
                    
                    for (int iC=0; iC < inImData[0].c(); iC++)
                        libIIPStable::fk_histogram_specification(rImData[0].v(iC), rImData[jj].v(iC), rImData[jj].v(iC),  rImData[0].w(),  rImData[0].h(), rImData[jj].w(), rImData[jj].h());
                    
                }
                
            }
            
            
            
            average_already_registered_single_burst(rImData, cfBurst/*s[ii]*/, fiNumber[ii], iAdaptiveBurst, iVerbose);            


            delete[] inImData;
            delete[] rImData;
            
            
            std::string filename(pout.value) ; filename += "_burst_";
            char numero[128]; sprintf(numero, "%d", ii);
            std::string snumero(numero); filename += snumero; filename += ".png";
                
            cfBurst.save(filename.c_str());
            

          
        }
        
        
        
    }
    
    
    
    
	    
    
    
}




