#include "libburst.h"




namespace libIIPStable 
{
    
    const float bg_here = -100.0f;
    
    
    
    
    //! Distance of borders of transformed image
    double max_dist_vertex(laMatrix &H, int iWidth, int iHeight)
	{
		
		double dMaxDist = 0.0f;
		
		
		// -- // check new position of vertexes of image // -- //
		for(int i = 0; i <= 1; i++)
			for(int j = 0; j <= 1; j++)
			{
				
				laVector vec(3), res(3);
				vec[0] = (double) (i*(iWidth)); 
				vec[1] = (double) (j*(iHeight)); 
				vec[2] = 1.0f; 
				
				res = H * vec;
                
				if (res[2] != 0.0f)
				{
					res[0] /= res[2]; res[1] /= res[2];
					
					double dDif = (res[0] - vec[0]) * (res[0] - vec[0]);
					dDif += (res[1] - vec[1]) * (res[1] - vec[1]);
					
					dDif = sqrt(dDif);
					if (dDif > dMaxDist) dMaxDist = dDif;
					
					
				} 
				
			}
		
		return dMaxDist;
		
	}
    
    
    
    
    //! Build groups with a prior subsampling
    void build_groups_of_images(cflimage *idata, int inImages, int &inBursts, int *fiIndexos, int *fiNumber, int iVerbose, int sfactor)
    {
        
        
        
		//--//   Parameters  // -- //
        
        //! sift parameters
        siftPar parameters;		
		default_sift_parameters(parameters);
        //parameters.PeakThresh = 2.0f;
		
        //! sift parameters, can be added in order to accelerate the matching part
        //parameters.MatchXradius = 0.25 * MAX(idata[0].w(), idata[0].h());
        //parameters.MatchYradius = 0.25 * MAX(idata[0].w(), idata[0].h());
        
        
        
		//! ransac parameters: given as input
		int iRansacIterarions = 1000;
		float fRansacTolerance = 1.50f;
		
        
		//!  minimum number of matches to compute an homography
		int inAccordedHom = 10;
        
        
        //! maximum distance par image corner allowed under homography transformation to be considered as a burst
		//! defined later since it depends on initial image size and not defined in here because of parallelism
        //float fBurstHomTol
        
        
        //! memory for homography
		laMatrix H(3,3), Hinv(3,3);
        
		
		//! memory for svd decomposition
		laVector Dsvd(2);
        laMatrix  Usvd(2,2), Vsvd(2,2), Asvd(2,2);
        
        
        //! init variables
        for (int ii=0; ii < inImages; ii++) fiIndexos[ii] = 0;
		fiIndexos[0] = 0;   // first burst begins at image 0
        fiNumber[0] = 1;    
		inBursts = 1;       
		
        
        
		
        
		//--// Dealing with first image //--//
        
        //! Subsample first image
		float fSigma;
        if (sfactor == 1) fSigma = 0.0f;
        else if (sfactor == 2) fSigma = 1.2;
        else if (sfactor == 3) fSigma = 1.69;
        else if (sfactor == 4) fSigma = 2.2;
        else
        {
            std::cout <<  "... subsampling factor must be lower than 5" << std::endl; exit(-1);
        }
        
        
        
        //! Convert to gray and apply SIFT
        cflimage init;
        if (sfactor == 1) init = idata[0];
        else  init = idata[0].subSample((float) sfactor, fSigma);
        
		flimage tmp_gray_init = init.getGray();     
		
        keypointslist keys0;                        
		compute_sift_keypoints(tmp_gray_init.v(), keys0, init.w(), init.h(), parameters);
		
		tmp_gray_init.erase();                             
		
		
		//! maximum distance par image corner allowed under homography transformation to be considered as a burst
		float fBurstHomTol = 0.10 * MAX(init.w(), init.h());
        
        
        //! print parameters
		if (iVerbose)
		{	
			std::cout << "\t... grouping bursts ( s = " << sfactor << ")" << std::endl;
		}
		
        
        
		//! begin process
		for(int iiIndex = 1; iiIndex < inImages; iiIndex++)
		{
			
			
			//--// Dealing with next image //--//		
		    
            //! subsample
			cflimage image;
            if (sfactor == 1) image = idata[iiIndex];
            else image =idata[iiIndex].subSample((float) sfactor, fSigma);
			
            
            //! convert to gray and apply SIFT
            flimage tmp_gray_image = image.getGray();						 
			keypointslist keys;											            
			compute_sift_keypoints(tmp_gray_image.v() ,keys, image.w(),image.h(),parameters);
			
			tmp_gray_image.erase();											
			
            
            //! compute sift matches
			matchingslist matchings;									
			compute_sift_matches(keys0,keys,matchings,parameters);
			
            
			
			//--//   RANSAC   //--//
			int n = matchings.size();										
			int naccorded=0;				
			if (n > inAccordedHom)											// if enough matches
			{
				
				float *x0 = new float[n]; float *y0 = new float[n]; 
				float *x1 = new float[n]; float *y1 = new float[n];
				
				matchingslist::iterator ptr = matchings.begin();			// store matches in float pointers
				for(int jj=0; jj < n; jj++, ptr++) 
				{
					x0[jj] = ptr->first.x;  y0[jj] = ptr->first.y;
					x1[jj] = ptr->second.x; y1[jj] = ptr->second.y;
				}
				
				
				int *accorded = new int[n];									// compute homography by applying Ransac
				
				naccorded = compute_ransac_planar_homography_n_points(x0,y0,x1,y1,n, iRansacIterarions, fRansacTolerance, H);
				
				delete[] x0; delete[] y0;
				delete[] x1; delete[] y1;
				delete[] accorded;
				
			} 
			
			
			
			
			//--//   Applying Homography   //--//
			if (iVerbose) std::cout << "\t(" << fiIndexos[inBursts-1] << " , " << iiIndex << " ):  " << n << " , " << naccorded  ; 
			
			
			//! Compute degrees of homography
			double h1,h2,dMaxDist,dTilt;
            
            
            if (naccorded > inAccordedHom)											
			{
                
                if (H[2][2] != 0.0f)										
                    for(int i=0; i < 3; i++) 
                        for(int j=0; j < 3; j++) H[i][j] /=  H[2][2];
                
                // h1, h2 to check if it is an affinity
                h1 = H[2][0];
                h2 = H[2][1];
                
                // distance of mapped corners
                dMaxDist = max_dist_vertex(H, init.w(), init.h());   
                
                // tilt
                Asvd[0][0]=H[0][0];
                Asvd[0][1]=H[0][1];
                Asvd[1][0]=H[1][0];
                Asvd[1][1]=H[1][1];
                
                compute_svd(Asvd, Usvd, Vsvd, Dsvd);
                
                dTilt = MAX(Dsvd[0],Dsvd[1]);
                dTilt /= MIN(Dsvd[0], Dsvd[1]);
			}
            
			
			
			if (naccorded > inAccordedHom && ((dMaxDist < fBurstHomTol && fabsf(h1) < 0.0001 & fabsf(h2) < 0.0001 && dTilt < 1.05)) )
			{
				
                if (iVerbose) std::cout << "  --> accepted (dMaxDist, h1, h2, Tilt) :  ("  << dMaxDist << " , " << h1 << " , " << h2 << " , " << dTilt << " )" << std::endl;
				
                //! update number of images in burst
                fiNumber[inBursts-1]++;
				
                
				
			}  else  //--//   It's not the same burst  //--//
			{
                
				if (iVerbose) std::cout << "  --> rejected (dMaxDist, h1, h2, Tilt) :  ("  << dMaxDist << " , " <<  h1 << " , " << h2 << " , " << dTilt << " )" << std::endl;
				
				fiIndexos[inBursts] = iiIndex;
				fiNumber[inBursts] = 1;
                
				keys0 = keys;
				
				inBursts++;
			}
			
			
		}
        
        
        
    }
    
    
    
    
    
    
    
    
    
    //! Build groups with a prior subsampling
    void build_groups_of_images_lm(char **idata, int inImages, int &inBursts, int *fiIndexos, int *fiNumber, int iVerbose, int sfactor)
    {
        
        
        
		//--//   Parameters  // -- //
        
        //! sift parameters
        siftPar parameters;		
		default_sift_parameters(parameters);
        //parameters.PeakThresh = 2.0f;
		
        //! sift parameters, can be added in order to accelerate the matching part
        //parameters.MatchXradius = 0.25 * MAX(idata[0].w(), idata[0].h());
        //parameters.MatchYradius = 0.25 * MAX(idata[0].w(), idata[0].h());
        
        
        
		//! ransac parameters: given as input
		int iRansacIterarions = 1000;
		float fRansacTolerance = 1.50f;
		
        
		//!  minimum number of matches to compute an homography
		int inAccordedHom = 10;
        
        
        //! maximum distance par image corner allowed under homography transformation to be considered as a burst
		//! defined later since it depends on initial image size and not defined in here because of parallelism
        //float fBurstHomTol
        
        
        //! memory for homography
		laMatrix H(3,3), Hinv(3,3);
        
		
		//! memory for svd decomposition
		laVector Dsvd(2);
        laMatrix  Usvd(2,2), Vsvd(2,2), Asvd(2,2);
        
        
        //! init variables
        for (int ii=0; ii < inImages; ii++) fiIndexos[ii] = 0;
		fiIndexos[0] = 0;   // first burst begins at image 0
        fiNumber[0] = 1;    
		inBursts = 1;       
		
        
        
		
        
		//--// Dealing with first image //--//
        
        //! Subsample first image
		float fSigma;
        if (sfactor == 1) fSigma = 0.0f;
        else if (sfactor == 2) fSigma = 1.2;
        else if (sfactor == 3) fSigma = 1.69;
        else if (sfactor == 4) fSigma = 2.2;
        else
        {
            std::cout <<  "... subsampling factor must be lower than 5" << std::endl; exit(-1);
        }
        
        
        
        //! Convert to gray and apply SIFT
        cflimage init;
        init.load(idata[0]);
        
        if (sfactor > 1)   init = init.subSample((float) sfactor, fSigma);
        
		flimage tmp_gray_init = init.getGray();     
		
        keypointslist keys0;                        
		compute_sift_keypoints(tmp_gray_init.v(), keys0, init.w(), init.h(), parameters);
		
		tmp_gray_init.erase();                             
		
		
		//! maximum distance par image corner allowed under homography transformation to be considered as a burst
		float fBurstHomTol = 0.10 * MAX(init.w(), init.h());
        
        
        //! print parameters
		if (iVerbose)
		{	
			std::cout << "\t... grouping bursts ( s = " << sfactor << ")" << std::endl;
		}
		
        
        
		//! begin process
		for(int iiIndex = 1; iiIndex < inImages; iiIndex++)
		{
			
			
			//--// Dealing with next image //--//		
		    
            //! subsample
			cflimage image;
            image.load(idata[iiIndex]);
            if (sfactor > 1) image = image.subSample((float) sfactor, fSigma);
			
            
            //! convert to gray and apply SIFT
            flimage tmp_gray_image = image.getGray();						 
			keypointslist keys;											            
			compute_sift_keypoints(tmp_gray_image.v() ,keys, image.w(),image.h(),parameters);
			
			tmp_gray_image.erase();											
			
            
            //! compute sift matches
			matchingslist matchings;									
			compute_sift_matches(keys0,keys,matchings,parameters);
			
            
			
			//--//   RANSAC   //--//
			int n = matchings.size();										
			int naccorded=0;				
			if (n > inAccordedHom)											// if enough matches
			{
				
				float *x0 = new float[n]; float *y0 = new float[n]; 
				float *x1 = new float[n]; float *y1 = new float[n];
				
				matchingslist::iterator ptr = matchings.begin();			// store matches in float pointers
				for(int jj=0; jj < n; jj++, ptr++) 
				{
					x0[jj] = ptr->first.x;  y0[jj] = ptr->first.y;
					x1[jj] = ptr->second.x; y1[jj] = ptr->second.y;
				}
				
				
				int *accorded = new int[n];									// compute homography by applying Ransac
				
				naccorded = compute_ransac_planar_homography_n_points(x0,y0,x1,y1,n, iRansacIterarions, fRansacTolerance, H);
				
				delete[] x0; delete[] y0;
				delete[] x1; delete[] y1;
				delete[] accorded;
				
			} 
			
			
			
			
			//--//   Applying Homography   //--//
			if (iVerbose) std::cout << "\t(" << fiIndexos[inBursts-1] << " , " << iiIndex << " ):  " << n << " , " << naccorded  ; 
			
			
			//! Compute degrees of homography
			double h1,h2,dMaxDist,dTilt;
            
            
            if (naccorded > inAccordedHom)											
			{
                
                if (H[2][2] != 0.0f)										
                    for(int i=0; i < 3; i++) 
                        for(int j=0; j < 3; j++) H[i][j] /=  H[2][2];
                
                // h1, h2 to check if it is an affinity
                h1 = H[2][0];
                h2 = H[2][1];
                
                // distance of mapped corners
                dMaxDist = max_dist_vertex(H, init.w(), init.h());   
                
                // tilt
                Asvd[0][0]=H[0][0];
                Asvd[0][1]=H[0][1];
                Asvd[1][0]=H[1][0];
                Asvd[1][1]=H[1][1];
                
                compute_svd(Asvd, Usvd, Vsvd, Dsvd);
                
                dTilt = MAX(Dsvd[0],Dsvd[1]);
                dTilt /= MIN(Dsvd[0], Dsvd[1]);
			}
            
			
			
			if (naccorded > inAccordedHom && ((dMaxDist < fBurstHomTol && fabsf(h1) < 0.0001 & fabsf(h2) < 0.0001 && dTilt < 1.05)) )
			{
				
                if (iVerbose) std::cout << "  --> accepted (dMaxDist, h1, h2, Tilt) :  ("  << dMaxDist << " , " << h1 << " , " << h2 << " , " << dTilt << " )" << std::endl;
				
                //! update number of images in burst
                fiNumber[inBursts-1]++;
				
                
				
			}  else  //--//   It's not the same burst  //--//
			{
                
				if (iVerbose) std::cout << "  --> rejected (dMaxDist, h1, h2, Tilt) :  ("  << dMaxDist << " , " <<  h1 << " , " << h2 << " , " << dTilt << " )" << std::endl;
				
				fiIndexos[inBursts] = iiIndex;
				fiNumber[inBursts] = 1;
                
				keys0 = keys;
				
				inBursts++;
			}
			
            
            image.erase();
            
		}
        
        
        
    }
    

    
    
    
    
    
    
    
    float compute_compression_factor(laMatrix &H, int  width, int height)
    {
        
        
        //! Compute baricenter of transformed points
        float bx = 0.0f, by = 0.0f;
        float num = 0.0f;
        for(int i = 0; i <= 1; i++)
            for(int j = 0; j <= 1; j++)
            {
                
                laVector vec(3), res(3);
                vec[0] = (float) (i*(width)); 
                vec[1] = (float) (j*(height)); 
                vec[2] = 1.0f; 
                
                res = H * vec;
                if (res[2] != 0.0f)  
                {
                    res[0] /= res[2]; res[1] /= res[2]; num++;
                    
                    bx += res[0];
                    by += res[1];
                    
                }
                
            }
        bx /= num; by /= num;
        
        
        
        //! Compute maximum distance of corners to baricenter
        float fDif1 = 0.0;
        for(int i = 0; i <= 1; i++)
            for(int j = 0; j <= 1; j++)
            {
                
                laVector vec(3), res(3);
                
                vec[0] = (float) (i*(width)); 
                vec[1] = (float) (j*(height)); 
                vec[2] = 1.0f; 
                
                res = H * vec;
                if (res[2] != 0.0f)  {
                    
                    res[0] /= res[2]; res[1] /= res[2];}
                
                
                float fDif = (res[0] - bx) * (res[0] - bx);
                fDif += (res[1] - by) * (res[1] - by);
                
                fDif = sqrtf(fDif);
                fDif1 = MAX(fDif1, fDif);
                
            }
        
        
        //! Compare to diagonal
        float fDiagonal = sqrtf((float) width * (float) width +  (float) height * (float) height);
        fDiagonal *= 0.5f;
        
        //float fDif0 = 0.25f * (float) width * (float) width + 0.25f  (float) height * (float) height;
        //fDif0 = sqrtf(fDif0);
        
        return fDif1 / fDiagonal;
        
        
    }
    
    
    
    
    /// Min singular value of the Jacobian of \a H at point (\a x,\a y).
    float minSVJacob(laMatrix & H, float x, float y)
    {
        
        
        laVector X(3);
        X[0] = x; X[1] = y; X[2] = 1.0f;
        
        laVector Y(3);
        Y = H*X;
        
        float d = 1.0 / Y[2];
        x *= d;
        y *= d;
        
        laMatrix J(2, 2);
        
        J[0][0] = d*(H[0][0]-H[2][0]*x);
        J[0][1] = d*(H[0][1]-H[2][1]*x);
        J[1][0] = d*(H[1][0]-H[2][0]*y);
        J[1][1] = d*(H[1][1]-H[2][1]*y);
        
        laMatrix U(2,2), V(2,2);
        laVector W(2);
        
        compute_svd(J, U, V, W);
        
        if (W[0] < W[1]) return W[0];
        else return W[1];
        
    }
    
    
    
    /// Compute min linear compression of \a H in rectangle [0,w]x[0,h].
    float minZoomOut(laMatrix &H, int w, int h)
    {
        float z=1.0;
        z = MAX(z, minSVJacob(H,0.0,0.0));
        z = MAX(z, minSVJacob(H,(float) w,0.0));
        z = MAX(z, minSVJacob(H,0.0,(float) h));
        z = MAX(z, minSVJacob(H,(float) w, (float) h));
        return z;
    }
    
    
    
    
    
    // creates an image mask of the Delaunay triangulation: fLarge inside the convex hull; bg_here outside
    void set_triangulation_mask(flimage &mask, struct triangulateio out)
    {
        int width = mask.w(), height = mask.h();
        
        
        // travel each pixel
        for(int ii=0; ii < width; ii++)
            for(int jj=0; jj < height; jj++)
            {
                float sign = 1; 
                mask.v(0)[width*jj+ii] = fLarge;
                
                // travel each edge of the convex hull
                for(int e = 0; e < out.numberofsegments; e++)
                {
                    // read vertices indexes
                    int e0_idx = out.segmentlist[e * 2];
                    int e1_idx = out.segmentlist[e * 2 + 1];
                    
                    // read vertices coordinates
                    float e0_x = out.pointlist[e0_idx * 2];
                    float e0_y = out.pointlist[e0_idx * 2 + 1];
                    float e1_x = out.pointlist[e1_idx * 2];
                    float e1_y = out.pointlist[e1_idx * 2 + 1];
                    
                    // compute director vector of the edge (always in the same order)
                    float v_x = e1_x - e0_x;
                    float v_y = e1_y - e0_y;
                    
                    // dot product of the normal vector to the edge with the point coordinates minus the first vertex
                    float product = -v_y*(ii - e0_x) + v_x*(jj - e0_y);
                    
                    // test sign of the dot product
                    if(e==0)
                    {
                        if(product >= 0) sign = 1;
                        else sign = -1;
                    }
                    else
                    {
                        // test sign
                        product = product*sign;
                        
                        // if sign changes set pixel image to zero and stop edge travelling
                        if(product < 0)
                        {
                            mask.v(0)[width*jj+ii] = bg_here;
                            e=out.numberofsegments;
                        }
                    }
                }
            }
        
    }
    
    
    
    
    
    
    void triangulate_Delaunay(float *x1, float *y1, int naccorded, flimage &mask)
    // void triangulate_Delaunay(float *x1, float *y1, int naccorded, cflimage &image, flimage &mask, int draw_borders, cflimage &borders)
    {
        // memory allocation and define input points
        struct triangulateio in, out;
        
        in.numberofpoints = naccorded;
        in.numberofpointattributes = 0;
        in.pointattributelist = (REAL *) NULL;//
        in.pointmarkerlist = (int *) NULL;//
        in.numberofsegments = 0;
        in.numberofholes = 0;
        in.numberofregions = 0;
        in.regionlist = (REAL *) NULL;//
        in.pointlist = (REAL *) malloc(in.numberofpoints * 2 * sizeof(REAL));
        for(int i=0; i<naccorded; i++)
        {
            in.pointlist[2*i] = x1[i];
            in.pointlist[2*i+1] = y1[i];
        }
        
        
        out.pointlist = (REAL *) NULL;            /* Not needed if -N switch used. */
        /* Not needed if -N switch used or number of attributes is zero: */
        out.pointattributelist = (REAL *) NULL;
        out.pointmarkerlist = (int *) NULL; /* Not needed if -N or -B switch used. */
        out.trianglelist = (int *) NULL;          /* Not needed if -E switch used. */
        /* Not needed if -E switch used or number of triangle attributes is zero: */
        out.triangleattributelist = (REAL *) NULL;
        /* Needed only if segments are output (-p or -c) and -P not used: */
        out.segmentlist = (int *) NULL;
        /* Needed only if segments are output (-p or -c) and -P and -B not used: */
        out.segmentmarkerlist = (int *) NULL;
        out.edgelist = (int *) NULL;             /* Needed only if -e switch used. */
        out.edgemarkerlist = (int *) NULL;   /* Needed if -e used and -B not used. */
        
        // triangulate
        triangulate("pcze", &in, &out, (struct triangulateio *) NULL);
        
        // write image with triangulation
        //draw_triangulation_image(image, out);
        
        // write borders of the convex hull of the triangulation
        //if(draw_borders) draw_convex_hull_borders(image, out, borders);
        
        // compute mask
        set_triangulation_mask(mask, out);
        
        // free memory
        free(in.pointlist);
        
    }
    
    
    
    
    
    
    
    
    
    // fit a polynomial of certain degree to the n matchings (x0,y0) <-> (x1,y1) 
    void fit_polynomial(float *x0, float *y0, float *x1, float *y1, float *coeff_x, float *coeff_y, int n, int degree, int width, int height)
    {
        // check degree
        int m;
        switch(degree)
        {
            case 2: m = 6;
                break;
            case 3: m = 10;
                break;
            case 4: m = 15;
                break;
            default: std::cout << "ERROR: Polynomial degree must be between 2 and 4. It has been fixed to 4." << std::endl;
                m = 15;
                break;
        }
        
        // built vectors coords0 and coords1 from (x0,y0)
        double *x = new double[n];
        double *y = new double[n];
        float cx = roundf((float) width/2.0f);
        float cy = roundf((float) height/2.0f);
        for(int i=0; i<n; i++)
        {
            x[i] = (double) ((x1[i]-cx)*2.0f/(float) width );
            y[i] = (double) ((y1[i]-cy)*2.0f/(float) height );
        }
        
        
        // built matrix X and X^t from (x1,y1)
        laMatrix X(n,m);
        laMatrix Xt(m,n);
        
        for(int i=0; i<n; i++)
        {
            int j = 0;
            for(int p=0; p <= degree; p++)
            {
                for(int q=0; q <= p; q++)
                {
                    
                    X[i][j] = (double) ( pow((x0[i]-cx)*2.0/ (double) width, (double (p-q))))  *  (double) ( pow((y0[i]-cy)*2.0/(double) height, (double) q) ) ;
                    
                    Xt[j][i] = X[i][j];
                    
                    j++;
                }
            }
        }
        
        // built matrix (X^t X)
        laMatrix XtX(m,m);
        for(int i=0; i<m; i++)
        {
            for(int j=0; j<m; j++)
            {
                double product = 0;
                for(int k=0; k<n; k++) product += Xt[i][k]*X[k][j];
                XtX[i][j] = product;
            }
        }
        
        // invert matrix (X^t X)
        laMatrix iXtX(m,m);
        luinv(XtX, iXtX);
        
        
        // compute vector of coefficients in x: inv(X^t X) X^t x
        double *Xtc = new double[m];
        for(int i=0; i<m; i++)
        {
            double product = 0;
            for(int j=0; j<n; j++) product += Xt[i][j]*x[j];
            Xtc[i] = product;//n;
        }
        for(int i=0; i<m; i++)
        {
            double product = 0;
            for(int j=0; j<m; j++) product += iXtX[i][j]*Xtc[j];
            coeff_x[i] =  (float) product;
        }
        
        // compute vector of coefficients in y: inv(X^t X) X^t y
        for(int i=0; i<m; i++)
        {
            double product = 0;
            for(int j=0; j<n; j++) product += Xt[i][j]*y[j];
            Xtc[i] = product;
        }
        for(int i=0; i<m; i++)
        {
            double product = 0;
            for(int j=0; j<m; j++) product += iXtX[i][j]*Xtc[j];
            coeff_y[i] = (float) product;
        }
        
        
        delete[] x;
        delete[] y;
        delete[] Xtc;
    }
    
    
    
    
    
    
    
    void float_vector_apply_polynomial(float x0s, float y0s, float *coeff_x, float *coeff_y, int width, int height, float *xp, float *yp)
    {
        int cx = round(width/2.0);
        int cy = round(height/2.0);
        *xp = 0;
        *yp = 0;
        int i = 0;
        for(int p=0; p <= 4; p++) // we assume that corresponding coefficients are zero if different polynomial degree
        {
            for(int q=0; q <= p; q++)
            {
                *xp += coeff_x[i]*pow((x0s-cx)*2.0/width,p-q)*pow((y0s-cy)*2.0/height,q);
                *yp += coeff_y[i]*pow((x0s-cx)*2.0/width,p-q)*pow((y0s-cy)*2.0/height,q);
                i++;
            }
        }
        *xp = (*xp)*width/2.0 + cx;
        *yp = (*yp)*height/2.0 + cy;
    }
    
    
    
    
    void apply_polynomial_picture(float *input, int width1, int height1, float *coeff_x, float *coeff_y, int spline_order, float *out, float x0, float y0, 
                                  int nwidth, int nheight, int zoom)
    {
        
        // splines stuff
        float *coeffs;
        float *ref;
        
        float  cx[12],cy[12],ak[13];
        
        if (spline_order!=0 && spline_order!=1 && spline_order!=-3 && 
            spline_order!=3 && spline_order!=5 && spline_order!=7 && spline_order!=9 && spline_order!=11)
        {	
            printf("unrecognized interpolation order.\n");
            exit(-1);
        }
        
        if (spline_order>=3) {
            coeffs = new float[width1*height1];
            finvsplineMW(input,spline_order,coeffs,width1,height1);
            ref = coeffs;
            if (spline_order>3) init_splinenMW(ak,spline_order);
            
        } else 
        {
            coeffs = NULL;
            ref = input;
        }
        
        float res;
        int n1,n2;
        float p=-0.5;
        
        // for each point in new image
        for(int i=0; i < nwidth; i++)
            for(int j=0; j < nheight; j++)
            {
                
                // position taking into account that a zoom is being applied
                float x0s, y0s;
                x0s = ((float) i / (float) zoom) + x0;
                y0s = ((float) j / (float) zoom) + y0;
                
                // compute position in image input
                float xp, yp;
                float_vector_apply_polynomial(x0s, y0s, coeff_x, coeff_y, width1, height1, &xp, &yp);
                
                int xi = (int)floor((double)xp); 
                int yi = (int)floor((double)yp);
                
                // interpolate values according to order and position with respect to the image borders		
                if(spline_order == 0) { 
					
                    if(xi<0 || xi>=width1 || yi<0 || yi>=height1)
                        res = bg_here; 
                    else res = input[yi*width1+xi];
					
                } else { 
                    
                    float bound=0;//3.0f;
                    if(xp<0.+bound || xp>=(float)width1-bound || yp<0.+bound || yp>=(float)height1-bound) res=bg_here; 
                    else 
                    {
                        //xp -= 0.5; yp -= 0.5;
                        float ux = xp-(float)xi;
                        float uy = yp-(float)yi;
						
                        switch(spline_order) 
                        {
                            case 1: /* first order interpolation (bilinear) */
                                n2 = 1;
                                cx[0]=ux; cx[1]=1.-ux;
                                cy[0]=uy; cy[1]=1.-uy;
                                break;
								
                            case -3: /* third order interpolation (bicubic Keys' function) */
                                n2 = 2;
                                keysMW(cx,ux,p);
                                keysMW(cy,uy,p);
                                break;
								
                            case 3: /* spline of order 3 */
                                n2 = 2;
                                spline3MW(cx,ux);
                                spline3MW(cy,uy);
                                break;
								
                            default: /* spline of order >3 */
                                n2 = (1+spline_order)/2;
                                splinenMW(cx,ux,ak,spline_order);
                                splinenMW(cy,uy,ak,spline_order);
                                break;
                        }
						
                        res = 0.; n1 = 1-n2;
                        
                        if(xi+n1>=0 && xi+n2<width1 && yi+n1>=0 && yi+n2<height1) {
							
                            int adr = yi*width1+xi; 
                            for(int dy=n1;dy<=n2;dy++) 
                                for(int dx=n1;dx<=n2;dx++) 
                                    res += cy[n2-dy]*cx[n2-dx]*ref[adr+width1*dy+dx];
                            
                        } else 
							
                            for(int dy=n1;dy<=n2;dy++)
                                for(int dx=n1;dx<=n2;dx++) 
                                    res += cy[n2-dy]*cx[n2-dx]*vMW(ref,xi+dx,yi+dy,bg_here,width1,height1);
						
                    }
					
                }	
				
                out[j*nwidth+i] = res;
                
            }
        
        if(coeffs != NULL) delete[] coeffs;
    }
    
    
    
    
	
	void register_burst_with_initial_image(cflimage *idata, cflimage *odata, int initIndex, int inImages, int iDistCorrectionFlag, int iDelaunayFlag, int iSiftType, int iVerbose, int & iCorrectlyRegistered, int *ipCorrectlyRegistered, int iCompressionFlag, float lScale, float uScale)
	{
		
        
		//--//   Parameters  // -- //
		
        
		//! Sift parameters
		siftPar parameters;		
		default_sift_parameters(parameters);
		
        
        //! SIFTIPOL / ASIFT parameters
        libSiftIPOLCODE::siftPar ASparameters;		
		libSiftIPOLCODE::default_sift_parameters(ASparameters);
		
        parameters.PeakThresh = 1.0f;
		
        
		
		//! Ransac parameters
		int iRansacIterations = 1000;
		float fRansacTolerance = 1.50f;
		
        
		//!  minimum number of matches to compute an homography
		int inAccordedHom = 10;
		
		
        //! print parameters
		if (iVerbose)
		{	
			std::cout << std::endl <<  "\t... registering groups " << std::endl;
		}
		
        
        
		
		//--// Dealing with first image //--//
		
        
        //! converting to gray and applying SIFT
		cflimage init = idata[initIndex];
		flimage tmp_gray_init = init.getGray();    		
		
		
        keypointslist keys0;   
        libSiftIPOLCODE::keypointslist ASkeys0;   
        
        
        if (iSiftType == 0)
            compute_sift_keypoints(tmp_gray_init.v(), keys0, init.w(), init.h(), parameters);
		else if (iSiftType == 1)
            libSiftIPOLCODE::compute_sift_keypoints(tmp_gray_init.v(), ASkeys0, init.w(), init.h(), ASparameters);
        
        
        
		tmp_gray_init.erase();                 
        
        
        
		//! Fill output first image
		odata[initIndex] = idata[initIndex];
		
		
		//! The first one is correctly registered with itself
        ipCorrectlyRegistered[0] = initIndex;
        iCorrectlyRegistered = 1;
        
        
        
        //omp_set_num_threads(6);
        
#ifdef _OPENMP
#pragma omp parallel shared(idata, odata, init, keys0)
#endif
		{
			
			
#ifdef _OPENMP
#pragma omp for schedule(dynamic) nowait 
#endif
			for(int iiIndex = initIndex + 1; iiIndex < initIndex + inImages; iiIndex++)
			{
				
                
				
				//--// Dealing with next image //--//		
				
                //! convert to gray and apply sift
				//! cflimage image=idata[iiIndex];
				flimage tmp_gray_image = idata[iiIndex].getGray();						 
				
				keypointslist keys;												 
				libSiftIPOLCODE::keypointslist ASkeys;												 
				
                if (iSiftType == 0)
                    compute_sift_keypoints(tmp_gray_image.v() ,keys, idata[iiIndex].w(),idata[iiIndex].h(),parameters);
                else if (iSiftType == 1)
                    libSiftIPOLCODE::compute_sift_keypoints(tmp_gray_image.v() , ASkeys, idata[iiIndex].w(),idata[iiIndex].h(), ASparameters);
                
                
				tmp_gray_image.erase();											 				
				
                //! compute sift matches
                matchingslist matchings;										 
                libSiftIPOLCODE::matchingslist ASmatchings;										 
				
                if (iSiftType == 0)
                    compute_sift_matches(keys0,keys,matchings,parameters);
                else if (iSiftType == 1)
                    libSiftIPOLCODE::compute_sift_matches(ASkeys0,ASkeys,ASmatchings,ASparameters);
                
                
				
				//--//   RANSAC   //--//
				
				int n = matchings.size();										
				int naccorded=0;				
				float *x0 = new float[n]; float *y0 = new float[n]; 
				float *x1 = new float[n]; float *y1 = new float[n];
				int *accorded = new int[n];
				
                // memory for homography
				laMatrix H(3,3), Hinv(3,3);
                
                
				if (n > inAccordedHom)											// if enough matches
				{
					
					matchingslist::iterator ptr = matchings.begin();			// store matches in float pointers
					libSiftIPOLCODE::matchingslist::iterator ASptr = ASmatchings.begin();			// store matches in float pointers
					
                    if (iSiftType == 0)
                    {
                        for(int jj=0; jj < n; jj++, ptr++) 
                        {
                            x0[jj] = ptr->first.x;  y0[jj] = ptr->first.y;
                            x1[jj] = ptr->second.x; y1[jj] = ptr->second.y;
                        }
					} else {
                        
                        for(int jj=0; jj < n; jj++, ptr++) 
                        {
                            x0[jj] = ptr->first.x;  y0[jj] = ASptr->first.y;
                            x1[jj] = ptr->second.x; y1[jj] = ASptr->second.y;
                        }
                        
                    }
					
                    
					naccorded = compute_ransac_planar_homography_n_points(x0,y0,x1,y1,n, iRansacIterations, fRansacTolerance, H, accorded);
					
					
				} 
				
				
                
                
                //--//   save keypoints and matches //--//
                /* int iSaveMatchesSiftFlag = 0;
                 if (iSaveMatchesSiftFlag)
                 {
                 
                 int widtho =  MAX(init.w(),image.w());
                 int heighto = init.h()+image.h();
                 flimage siftOut(widtho, heighto);
                 
                 for(int jj = 0; jj < init.h(); jj++)
                 for(int ii = 0; ii < init.w(); ii++)  siftOut[jj*widtho+ii] = tmp_gray_init[jj*init.w()+ii];	
                 
                 for(int jj = 0; jj < image.h(); jj++)
                 for(int ii = 0; ii < image.w(); ii++)  siftOut[(init.h() + jj)*widtho + ii] = tmp_gray_image[jj * image.w() + ii];	
                 
                 for (int ii=0; ii < naccorded; ii++)
                 {
                 
                 fiImageDrawLine(siftOut.v(), (int) x0[ accorded[ii] ], (int) y0[ accorded[ii] ], (int) x1[ accorded[ii] ], (int) y1[ accorded[ii] ]+ init.h(), 255.0f, widtho, heighto);
                 }
                 
                 
                 std::string filename; filename = "matches";
                 char numero[128]; sprintf(numero, "%d", initIndex);
                 std::string snumero(numero); filename += snumero;
                 
                 filename += "_";
                 
                 char numero2[128]; sprintf(numero2, "%d", iiIndex);
                 std::string snumero2(numero2); filename += snumero2;
                 
                 filename += ".png";
                 
                 siftOut.save(filename.c_str());
                 
                 }
                 */
                
                
                
				
				//--//   applying homography   //--//
				if (iVerbose) std::cout << "\t(0 , " << iiIndex << " ):  " << n << " , " << naccorded  ; 
				
				
				//! normalize homography
				if (H[2][2] != 0.0f)										
					for(int i=0; i < 3; i++) 
						for(int j=0; j < 3; j++) H[i][j] /=  H[2][2];
				
				
                //! estimate compression factor 
                float fCompresion = 1.0f / compute_compression_factor(H, idata[iiIndex].w(), idata[iiIndex].h());
                if (iVerbose) { std::cout << " (compression: "<<  fCompresion << ")" ; }
                
                
                //! alternative way of computing compression, gets identical results
                //float fCompresion2 =  1.0f / minZoomOut(H, image.w(), image.h());
                
                
                //! Apply homography if enough matches to accept transformation + correct compression if flag activated + no distorsion flag
				if (naccorded > inAccordedHom && !iDistCorrectionFlag && (!iCompressionFlag  ||  (iCompressionFlag && fCompresion >= lScale && fCompresion <= uScale) ) )
				{
                    
                    
                    //! Increment index of correctly registered frames
                    ipCorrectlyRegistered[iCorrectlyRegistered] = iiIndex;
                    iCorrectlyRegistered++;
                    
                    if (iVerbose) { std::cout << "\t -->  registered"<< std::endl; }
					
                    
					//! apply transormation tmp = H-1 (image)
					luinv(H, Hinv);									
					
                    
					//! transformed image has same size than initial
					odata[iiIndex].create(init.w(), init.h(), init.c());
					odata[iiIndex] = BagValueAfterRegistration;
					
					
					//! apply homography first in order to avoid boundary effects
					flimage tmpMaskAux(idata[iiIndex].w(), idata[iiIndex].h());			
					flimage tmpMask(init.w(), init.h());
					
                    //! we use the image itself as mask, since it can be already registered and only a part be of interest
					tmpMaskAux = idata[iiIndex].getGray();
					tmpMask = BagValueAfterRegistration;
					
					apply_planar_homography(tmpMaskAux.v(),tmpMaskAux.w(),tmpMaskAux.h(), Hinv, BagValueAfterRegistration, HOMOGRAPHY_RESAMPLING_ORDER,tmpMask.v(),0.0,0.0,tmpMask.w(),tmpMask.h());				
					
                    //! resize tmpMaskAux and use for erosion
					tmpMaskAux = tmpMask;
                    fiPatchMin(tmpMaskAux.v(), tmpMask.v(), 3.0, tmpMask.w(), tmpMask.h());
                    
					
                    //! tmpMask contains the valid points to be kept after homography
					//! Transform images
					for (int ii=0; ii < idata[iiIndex].c(); ii++)
					{
						apply_planar_homography(idata[iiIndex].v(ii),idata[iiIndex].w(),idata[iiIndex].h(), Hinv, BagValueAfterRegistration, HOMOGRAPHY_RESAMPLING_ORDER, odata[iiIndex].v(ii),0.0,0.0,odata[iiIndex].w(),odata[iiIndex].h());
						
						for (int ijk=0; ijk < odata[iiIndex].wh(); ijk++) 
						{
							if (tmpMask[ijk] < BagValueAfterRegistration/2.0) odata[iiIndex][ii * odata[iiIndex].wh() + ijk] = BagValueAfterRegistration;
							
						}
						
					}
					
					
                    //! Apply distorsion correction if enough matches to accept transformation + correct compression if flag activated 
				} else if (naccorded > inAccordedHom  && (!iCompressionFlag  ||  (iCompressionFlag && fCompresion >= lScale && fCompresion <= uScale) ))
				{
					
                    
                    //! Increment index of correctly registered frames
                    ipCorrectlyRegistered[iCorrectlyRegistered] = iiIndex;
                    iCorrectlyRegistered++;
                    
                    
                    if (iVerbose) { std::cout << "\t -->  registered" << std::endl; }
                    
					// Adjust polynomial to inliers
					float *x0i = new float[naccorded];
					float *y0i = new float[naccorded];
					float *x1i = new float[naccorded];
					float *y1i = new float[naccorded];
					
					for(int i=0; i<naccorded; i++)
					{
						x0i[i] = x0[accorded[i]];
						y0i[i] = y0[accorded[i]];
						x1i[i] = x1[accorded[i]];
						y1i[i] = y1[accorded[i]];
					}
					
					
					int degree = 4;
					float *coeff_x = new float[15];
					float *coeff_y = new float[15];
					
					fit_polynomial(x0i, y0i, x1i, y1i, coeff_x, coeff_y, naccorded, degree, init.w(), init.h());
					
                    
                    
                    //! Adjust delaunay triangulation with x0, y0
                    flimage dlMask(idata[iiIndex].w(), idata[iiIndex].h());
                    if (iDelaunayFlag) triangulate_Delaunay(x1i, y1i, naccorded, dlMask);
                    
                    
                    
                    
					
					//! transformed image has same size than initial
					odata[iiIndex].create(init.w(), init.h(), init.c());
					odata[iiIndex] = BagValueAfterRegistration;
					
					
					//! use mask to avoid boundary effects and apply erosion
					flimage tmpMaskAux(idata[iiIndex].w(), idata[iiIndex].h());			
					flimage tmpMask(init.w(), init.h());
					
                    //! Use previous image, since this can be a reregistration and points outside have value -1.0f;
					tmpMaskAux = idata[iiIndex].getGray();
                    
                    
                    //! Use minimum with convex hull with is bg_here outside hull
                    if (iDelaunayFlag)
                    {
                        for (int iikk=0; iikk < tmpMaskAux.wh(); iikk++)
                        {
                            tmpMaskAux[iikk] = MIN(tmpMaskAux[iikk], dlMask[iikk]);
                        }
                    } 
                    
                 
                    //! free space for dlMask
                    dlMask.erase();
                    
                   					
                
					tmpMask = BagValueAfterRegistration;
					
					
					apply_polynomial_picture(tmpMaskAux.v(),tmpMaskAux.w(),tmpMaskAux.h(), coeff_x, coeff_y,  POLYNOMIAL_RESAMPLING_ORDER, tmpMask.v(),0.0,0.0,tmpMask.w(),tmpMask.h(), 1); // Zoom:1
					
					
					tmpMaskAux = tmpMask;
					fiPatchMin(tmpMaskAux.v(), tmpMask.v(), 3.0, tmpMask.w(), tmpMask.h());
                    
                    
                    //! free space for tmpMaskAux 
                    tmpMaskAux.erase();
                    
                    

              
                    
                    
					//! Transform images
					for (int ii=0; ii < idata[iiIndex].c(); ii++)
					{
						
						apply_polynomial_picture(idata[iiIndex].v(ii),idata[iiIndex].w(),idata[iiIndex].h(), coeff_x,coeff_y,  POLYNOMIAL_RESAMPLING_ORDER, odata[iiIndex].v(ii),0.0,0.0,odata[iiIndex].w(),odata[iiIndex].h(), 1); // Zoom:1
						
						for (int ijk=0; ijk < odata[iiIndex].wh(); ijk++) 
						{
							if (tmpMask[ijk] < BagValueAfterRegistration/2) odata[iiIndex][ii * odata[iiIndex].wh() + ijk] = BagValueAfterRegistration;
							
                       
                            
						}
						
					}
					
                    
                    
                    //! free space for tmpMask
                    tmpMask.erase();
                    
                    
                    
                    //! Take into account convex hull in the first image
                    
                    flimage dlMask0(init.w(), init.h());
                    if (iDelaunayFlag) triangulate_Delaunay(x0i, y0i, naccorded, dlMask0);

                    
                    for (int ii=0; ii < idata[iiIndex].c(); ii++)
					{
						for (int ijk=0; ijk < odata[iiIndex].wh(); ijk++) 
						{
							
							
                            if (dlMask0[ijk] < 0.0f) odata[iiIndex][ii * odata[iiIndex].wh() + ijk] = BagValueAfterRegistration;
                            
                            
						}
					}
                    
                    
                    //! free space for dlMask0
                    dlMask0.erase();
                    
                    
                    delete[] x0i;
					delete[] y0i;
					delete[] x1i;
					delete[] y1i;

                    
                    
                    delete[] coeff_x;
                    delete[] coeff_y;

					
				} else
                {
                    
                    if (iVerbose) { std::cout << "\t -->  not registered"<< std::endl; }
                    odata[iiIndex] = idata[iiIndex];
                    
                }
				
				
				
				
				delete[] x0; delete[] y0;
				delete[] x1; delete[] y1;
				delete[] accorded;
				
				
			} 
			
		}
	}
	
    
    
#if 0
    void register_burst_with_initial_image_zoom(cflimage *idata, cflimage *odata, int initIndex, int inImages, int iDistCorrectionFlag, int iDelaunayFlag, int iSiftType, int iVerbose, int & iCorrectlyRegistered, int *ipCorrectlyRegistered, int iCompressionFlag, float lScale, float uScale, float fZoom)
	{
		
        
		//--//   Parameters  // -- //
		
        
		//! Sift parameters
		siftPar parameters;		
		default_sift_parameters(parameters);
		
        
        //! SIFTIPOL / ASIFT parameters
        libSiftIPOLCODE::siftPar ASparameters;		
		libSiftIPOLCODE::default_sift_parameters(ASparameters);
		
        parameters.PeakThresh = 1.0f;
		
        
		
		//! Ransac parameters
		int iRansacIterations = 1000;
		float fRansacTolerance = 1.50f;
		
        
		//!  minimum number of matches to compute an homography
		int inAccordedHom = 10;
		
		
        //! print parameters
		if (iVerbose)
		{	
			std::cout << std::endl <<  "\t... registering groups " << std::endl;
		}
		
        
        
		
		//--// Dealing with first image //--//
		
        
        //! converting to gray and applying SIFT
		cflimage init = idata[initIndex];
		flimage tmp_gray_init = init.getGray();    		
		
		
        keypointslist keys0;   
        libSiftIPOLCODE::keypointslist ASkeys0;   
        
        
        if (iSiftType == 0)
            compute_sift_keypoints(tmp_gray_init.v(), keys0, init.w(), init.h(), parameters);
		else if (iSiftType == 1)
            libSiftIPOLCODE::compute_sift_keypoints(tmp_gray_init.v(), ASkeys0, init.w(), init.h(), ASparameters);
        
        
        
		tmp_gray_init.erase();                 
        
        
        
		//! Fill output first image
        int splinesOrder=3;
		if (fZoom > 1.0)
            odata[initIndex] = idata[initIndex].upSampleSplines(fZoom, splinesOrder);
        else
            odata[initIndex] = idata[initIndex];
		
		
        
		//! The first one is correctly registered with itself
        ipCorrectlyRegistered[0] = initIndex;
        iCorrectlyRegistered = 1;
        
        
        
         omp_set_num_threads(6);
        
#pragma omp parallel shared(idata, odata, init, keys0)
		{
			
			
#pragma omp for schedule(dynamic) nowait 
			for(int iiIndex = initIndex + 1; iiIndex < initIndex + inImages; iiIndex++)
			{
				
                
				
				//--// Dealing with next image //--//		
				
                //! convert to gray and apply sift
				//cflimage image=idata[iiIndex];
				flimage tmp_gray_image = idata[iiIndex].getGray();						 
				
				keypointslist keys;												 
				libSiftIPOLCODE::keypointslist ASkeys;												 
				
                if (iSiftType == 0)
                    compute_sift_keypoints(tmp_gray_image.v() ,keys, idata[iiIndex].w(),idata[iiIndex].h(),parameters);
                else if (iSiftType == 1)
                    libSiftIPOLCODE::compute_sift_keypoints(tmp_gray_image.v() , ASkeys, idata[iiIndex].w(),idata[iiIndex].h(), ASparameters);
                
                
				tmp_gray_image.erase();											 				
				
                //! compute sift matches
                matchingslist matchings;										 
                libSiftIPOLCODE::matchingslist ASmatchings;										 
				
                if (iSiftType == 0)
                    compute_sift_matches(keys0,keys,matchings,parameters);
                else if (iSiftType == 1)
                    libSiftIPOLCODE::compute_sift_matches(ASkeys0,ASkeys,ASmatchings,ASparameters);
                
                
				
				//--//   RANSAC   //--//
				
				int n = matchings.size();										
				int naccorded=0;				
				float *x0 = new float[n]; float *y0 = new float[n]; 
				float *x1 = new float[n]; float *y1 = new float[n];
				int *accorded = new int[n];
				
                // memory for homography
				laMatrix H(3,3), Hinv(3,3);
                
                
				if (n > inAccordedHom)											// if enough matches
				{
					
					matchingslist::iterator ptr = matchings.begin();			// store matches in float pointers
					libSiftIPOLCODE::matchingslist::iterator ASptr = ASmatchings.begin();			// store matches in float pointers
					
                    if (iSiftType == 0)
                    {
                        for(int jj=0; jj < n; jj++, ptr++) 
                        {
                            x0[jj] = ptr->first.x;  y0[jj] = ptr->first.y;
                            x1[jj] = ptr->second.x; y1[jj] = ptr->second.y;
                        }
					} else {
                        
                        for(int jj=0; jj < n; jj++, ptr++) 
                        {
                            x0[jj] = ptr->first.x;  y0[jj] = ASptr->first.y;
                            x1[jj] = ptr->second.x; y1[jj] = ASptr->second.y;
                        }
                        
                    }
					
                    
					naccorded = compute_ransac_planar_homography_n_points(x0,y0,x1,y1,n, iRansacIterations, fRansacTolerance, H, accorded);
					
					
				} 
				

                
                
                
                
				
				//--//   applying homography   //--//
				if (iVerbose) std::cout << "\t(0 , " << iiIndex << " ):  " << n << " , " << naccorded  ; 
				
				
				//! normalize homography
				if (H[2][2] != 0.0f)										
					for(int i=0; i < 3; i++) 
						for(int j=0; j < 3; j++) H[i][j] /=  H[2][2];
				
				
                //! estimate compression factor 
                float fCompresion = 1.0f / compute_compression_factor(H, idata[iiIndex].w(), idata[iiIndex].h());
                if (iVerbose) { std::cout << " (compression: "<<  fCompresion << ")" ; }
                
                
                //! alternative way of computing compression, gets identical results
                //float fCompresion2 =  1.0f / minZoomOut(H, image.w(), image.h());
                
                
                //! Apply homography if enough matches to accept transformation + correct compression if flag activated + no distorsion flag
				if (naccorded > inAccordedHom && !iDistCorrectionFlag && (!iCompressionFlag  ||  (iCompressionFlag && fCompresion >= lScale && fCompresion <= uScale) ) )
				{
                    
                    
                    //! Increment index of correctly registered frames
                    ipCorrectlyRegistered[iCorrectlyRegistered] = iiIndex;
                    iCorrectlyRegistered++;
                    
                    if (iVerbose) { std::cout << "\t -->  registered"<< std::endl; }
					
                    
					//! apply transormation tmp = H-1 (image)
					luinv(H, Hinv);									
					

                    //! OJO arrived here
					//! transformed image has  size of initial multiplied by fZoom
					odata[iiIndex].create(odata[initIndex].w(), odata[initIndex].h(), odata[initIndex].c());
					odata[iiIndex] = BagValueAfterRegistration;
					
					
					//! apply homography first in order to avoid boundary effects
					flimage tmpMaskAux(idata[iiIndex].w(), idata[iiIndex].h());			
					flimage tmpMask(odata[initIndex].w(), odata[initIndex].h());
					
                    //! we use the image itself as mask, since it can be already registered and only a part be of interest
					tmpMaskAux = idata[iiIndex].getGray();
					tmpMask = BagValueAfterRegistration;
					
                    
					apply_planar_homography_zoom(tmpMaskAux.v(),tmpMaskAux.w(),tmpMaskAux.h(), Hinv, BagValueAfterRegistration, HOMOGRAPHY_RESAMPLING_ORDER,tmpMask.v(),0.0,0.0,tmpMask.w(),tmpMask.h(), fZoom);				
					
                    
                    //! resize tmpMaskAux and use for erosion
					tmpMaskAux = tmpMask;
                    fiPatchMin(tmpMaskAux.v(), tmpMask.v(), 3.0, tmpMask.w(), tmpMask.h());
                    
					
                    //! tmpMask contains the valid points to be kept after homography
					//! Transform images
					for (int ii=0; ii < idata[iiIndex].c(); ii++)
					{
						apply_planar_homography_zoom(idata[iiIndex].v(ii),idata[iiIndex].w(),idata[iiIndex].h(), Hinv, BagValueAfterRegistration, HOMOGRAPHY_RESAMPLING_ORDER, odata[iiIndex].v(ii),0.0,0.0,odata[iiIndex].w(),odata[iiIndex].h(), fZoom);
						
						for (int ijk=0; ijk < odata[iiIndex].wh(); ijk++) 
						{
							if (tmpMask[ijk] < 0.0f) odata[iiIndex][ii * odata[iiIndex].wh() + ijk] = BagValueAfterRegistration;
							
						}
						
					}
					
					
                    //! Apply distorsion correction if enough matches to accept transformation + correct compression if flag activated 
				} else if (naccorded > inAccordedHom  && (!iCompressionFlag  ||  (iCompressionFlag && fCompresion >= lScale && fCompresion <= uScale) ))
				{
				/*	
                    
                    //! Increment index of correctly registered frames
                    ipCorrectlyRegistered[iCorrectlyRegistered] = iiIndex;
                    iCorrectlyRegistered++;
                    
                    
                    if (iVerbose) { std::cout << "\t -->  registered" << std::endl; }
                    
					// Adjust polynomial to inliers
					float *x0i = new float[naccorded];
					float *y0i = new float[naccorded];
					float *x1i = new float[naccorded];
					float *y1i = new float[naccorded];
					
					for(int i=0; i<naccorded; i++)
					{
						x0i[i] = x0[accorded[i]];
						y0i[i] = y0[accorded[i]];
						x1i[i] = x1[accorded[i]];
						y1i[i] = y1[accorded[i]];
					}
					
					
					int degree = 4;
					float *coeff_x = new float[15];
					float *coeff_y = new float[15];
					
					fit_polynomial(x0i, y0i, x1i, y1i, coeff_x, coeff_y, naccorded, degree, init.w(), init.h());
					
                    
                    
                    //! Adjust delaunay triangulation
                    flimage dlMask(image.w(), image.h());
                    flimage dlMask0(init.w(), init.h());
                    
                    if (iDelaunayFlag) triangulate_Delaunay(x1i, y1i, naccorded, dlMask);
                    if (iDelaunayFlag) triangulate_Delaunay(x0i, y0i, naccorded, dlMask0);
                    
                    
                    
					delete[] x0i;
					delete[] y0i;
					delete[] x1i;
					delete[] y1i;
					
                    
					
					//! transformed image has same size than initial
					odata[iiIndex].create(init.w(), init.h(), init.c());
					odata[iiIndex] = BagValueAfterRegistration;
					
					
					//! use mask to avoid boundary effects and apply erosion
					flimage tmpMaskAux(image.w(), image.h());			
					flimage tmpMask(init.w(), init.h());
					
                    //! Use previous image, since this can be a reregistration and points outside have value -1.0f;
					tmpMaskAux = image.getGray();
                    
                    
                    //! Use minimum with convex hull with is bg_here outside hull
                    if (iDelaunayFlag)
                    {
                        for (int iikk=0; iikk < tmpMaskAux.wh(); iikk++)
                        {
                            tmpMaskAux[iikk] = MIN(tmpMaskAux[iikk], dlMask[iikk]);
                        }
                    } 
                    
                    
					tmpMask = BagValueAfterRegistration;
					
					
					apply_polynomial_picture(tmpMaskAux.v(),tmpMaskAux.w(),tmpMaskAux.h(), coeff_x, coeff_y,  POLYNOMIAL_RESAMPLING_ORDER, tmpMask.v(),0.0,0.0,tmpMask.w(),tmpMask.h(), 1); // Zoom:1
					
					
					tmpMaskAux = tmpMask;
					fiPatchMin(tmpMaskAux.v(), tmpMask.v(), 3.0, tmpMask.w(), tmpMask.h());
                    
                    
                    
					//! Transform images
					for (int ii=0; ii < image.c(); ii++)
					{
						
						apply_polynomial_picture(image.v(ii),image.w(),image.h(), coeff_x,coeff_y,  POLYNOMIAL_RESAMPLING_ORDER, odata[iiIndex].v(ii),0.0,0.0,odata[iiIndex].w(),odata[iiIndex].h(), 1); // Zoom:1
						
						for (int ijk=0; ijk < odata[iiIndex].wh(); ijk++) 
						{
							if (tmpMask[ijk] < 0.0f) odata[iiIndex][ii * odata[iiIndex].wh() + ijk] = BagValueAfterRegistration;
							
                            if (dlMask0[ijk] < 0.0f) odata[iiIndex][ii * odata[iiIndex].wh() + ijk] = BagValueAfterRegistration;
                            
                            
						}
						
					}
					
				*/	
				} else
                {
                    
                    if (iVerbose) { std::cout << "\t -->  not registered"<< std::endl; }
                    odata[iiIndex] = idata[iiIndex];
                    
                }
				
				
				
				
				delete[] x0; delete[] y0;
				delete[] x1; delete[] y1;
				delete[] accorded;
				
				
				
			} 
			
		}
	}
#endif
	
    
    
    
    
    
    
    
    void  average_already_registered_burst(cflimage *idata, cflimage *cfBursts, int inBursts, int * fiIndex, int inImages, int iAdaptiveBurst, int iVerbose)
    {
		
		
        
#ifdef _OPENMP
#pragma omp parallel shared(idata,cfBursts,inBursts,fiIndex)
#endif
        {
            
            
#ifdef _OPENMP
#pragma omp for schedule(dynamic) nowait
#endif
            for (int ii=0; ii < inBursts; ii++)
            {
                
                //! Create fused image
                cfBursts[ii].create(idata[fiIndex[ii]].w(), idata[fiIndex[ii]].h(), idata[fiIndex[ii]].c() );
                cfBursts[ii] = 0.0f;
                
                
                //! Create counter
                flimage fpCounter(cfBursts[ii].w(), cfBursts[ii].h()); fpCounter=0.0f;
                
                
                for (int jj = fiIndex[ii]; (ii != inBursts-1 && jj < fiIndex[ii+1]) || (ii == inBursts-1 && jj < inImages); jj++) 
                {
                    
                    
                    //! Create weight image
                    flimage fpWeight(cfBursts[ii].w(), cfBursts[ii].h());  fpWeight=1.0f;
                    
                    
                    
                    
                    if (iAdaptiveBurst)
                    {
                        
                        int Diri_halfBlockSize = 50;
                        
                        //! Create gradient, counter and weight image
                        flimage fpGray(cfBursts[ii].w(), cfBursts[ii].h());
                        flimage fpGradient(cfBursts[ii].w(), cfBursts[ii].h());
                        
                        
                        
                        fpGray = idata[jj].getGray();
                        //! Attention, since images are transformed, we could have large a gradient at boundary of transformed image
                        //! or large zones with gradient 0
                        fpGradient = fpGray.gradient('f');	
                        
                        
                        //! We use |grad_u| instad of sqrt(|grad_u|^2 - sigma^2)
                        //! for (int kk = 0; kk < fpGradient.wh(); kk++) fpGradient[kk] = fpGradient[kk] * fpGradient[kk];
                        
                        
                        
                        
                        
                        //! Integral image
                        flimage gradIntegralImage(fpGradient.w(), fpGradient.h()); gradIntegralImage = 0.0f;
                        fiComputeIntegralImage(fpGradient.v(), gradIntegralImage.v(), fpGradient.w(), fpGradient.h());
                        
                        fpWeight=0.0f;
                        for(int pi=0; pi < gradIntegralImage.w(); pi++)
                            for(int pj=0; pj < gradIntegralImage.h(); pj++)
                            {
                                
                                //! We do not care about boundary at the moment
                                //! compute the Dirichlet integral in a block centered at the pixel
                                //! the integral on a block with corners 1,2,3,4 (corner ordering from top to bottom and from left to right) is
                                //! integral(4) + integral(1) - integral(3) - integral(2)
                                
                                int di_p = Diri_halfBlockSize, di_m = Diri_halfBlockSize;
                                int dj_p = Diri_halfBlockSize, dj_m = Diri_halfBlockSize;
                                
                                if(pi+di_p >= gradIntegralImage.w()) di_p = gradIntegralImage.w() - 1 - pi;
                                if(pj+dj_p >= gradIntegralImage.h()) dj_p = gradIntegralImage.h() - 1 - pj;
                                if(pi-di_m < 0) di_m = pi;
                                if(pj-dj_m < 0) dj_m = pj;
                                
                                int ipij = pj * fpWeight.w() + pi;
                                fpWeight[ipij] = gradIntegralImage[fpWeight.w()*(pj+dj_p)+(pi+di_p)];
                                fpWeight[ipij] += gradIntegralImage[fpWeight.w()*(pj-dj_m)+(pi-di_m)];
                                fpWeight[ipij] -= gradIntegralImage[fpWeight.w()*(pj+dj_p)+(pi-di_m)];
                                fpWeight[ipij] -= gradIntegralImage[fpWeight.w()*(pj-dj_m)+(pi+di_p)];
                                
                                
                                //! We use |grad_u| instad of sqrt(|grad_u|^2 - sigma^2)
                                //! fpWeight[ipij] = sqrtf(fpWeight[ipij]);
                                
                            }
                        
                        
                        
                        
                        
                        //! Slow version
                        //fpWeight = fpGradient.patchMean(Diri_halfBlockSize);
                        
                        
                        //! avoid problems for points without any non zero weight
                        for (int kk = 0; kk < fpWeight.wh(); kk++) fpWeight[kk] = MAX(fpWeight[kk], 0.001f);
                        
                        
                    }
                    
                    
                    
                    //! accumulate transformed images
                    for(int kk = 0; kk < cfBursts[ii].c(); kk++)
                    {
                        float *igray =  cfBursts[ii].v(kk);
                        float *tgray  = idata[jj].v(kk);
                        
                        for (int ll=0; ll < idata[jj].wh(); ll++)
                            if (tgray[ll] > BagValueAfterRegistration)
                                igray[ll] +=  fpWeight[ll] * tgray[ll];
                        
                    }
                    
                    
                    //! accumulate counter
                    float *tgray  =  idata[jj].v(0);
                    for (int ll=0; ll < idata[jj].wh(); ll++)
                        if (tgray[ll] > BagValueAfterRegistration)
				fpCounter[ll] += fpWeight[ll];  					
                    
                    
                    
                }
                
                
                
                
                
                
                
                //! normalize burst image
                for(int kk = 0; kk < cfBursts[ii].c(); kk++)
                {
                    float *igray =  cfBursts[ii].v(kk);   
                    for (int ll=0; ll < cfBursts[ii].wh(); ll++)
			    if (fpCounter[ll] > 0.0f)
				    igray[ll] /= fpCounter[ll];
			    else
				    fprintf(stderr, "Warning: dividing by zero"
				    " when averaging registered images\n");
                    
                }
                
            }
            
            
            
        }
        
        
	}
	
    
    
    
    
    
    
    
    
    void  average_already_registered_single_burst(cflimage *idata, cflimage &cfBursts, int inImages, int iAdaptiveBurst, int iVerbose)
    {
		
 
                //! Create fused image
                cfBursts.erase();
                cfBursts.create(idata[0].w(), idata[0].h(), idata[0].c() );
                cfBursts = 0.0f;
                
                
                //! Create counter
                flimage fpCounter(cfBursts.w(), cfBursts.h()); fpCounter=0.0f;
                
                
                for (int jj = 0; jj < inImages; jj++) 
                {
                    
                    
                    //! Create weight image
                    flimage fpWeight(cfBursts.w(), cfBursts.h());  fpWeight=1.0f;
                    
                    
                    
                    
                    if (iAdaptiveBurst)
                    {
                        
                        int Diri_halfBlockSize = 50;
                        
                        //! Create gradient, counter and weight image
                        flimage fpGray(cfBursts.w(), cfBursts.h());
                        flimage fpGradient(cfBursts.w(), cfBursts.h());
                        
                        
                        
                        fpGray = idata[jj].getGray();
                        //! Attention, since images are transformed, we could have large a gradient at boundary of transformed image
                        //! or large zones with gradient 0
                        fpGradient = fpGray.gradient('f');	
                        
                        
                        //! We use |grad_u| instad of sqrt(|grad_u|^2 - sigma^2)
                        //! for (int kk = 0; kk < fpGradient.wh(); kk++) fpGradient[kk] = fpGradient[kk] * fpGradient[kk];
                        
                        
                        
                        
                        
                        //! Integral image
                        flimage gradIntegralImage(fpGradient.w(), fpGradient.h()); gradIntegralImage = 0.0f;
                        fiComputeIntegralImage(fpGradient.v(), gradIntegralImage.v(), fpGradient.w(), fpGradient.h());
                        
                        fpWeight=0.0f;
                        for(int pi=0; pi < gradIntegralImage.w(); pi++)
                            for(int pj=0; pj < gradIntegralImage.h(); pj++)
                            {
                                
                                //! We do not care about boundary at the moment
                                //! compute the Dirichlet integral in a block centered at the pixel
                                //! the integral on a block with corners 1,2,3,4 (corner ordering from top to bottom and from left to right) is
                                //! integral(4) + integral(1) - integral(3) - integral(2)
                                
                                int di_p = Diri_halfBlockSize, di_m = Diri_halfBlockSize;
                                int dj_p = Diri_halfBlockSize, dj_m = Diri_halfBlockSize;
                                
                                if(pi+di_p >= gradIntegralImage.w()) di_p = gradIntegralImage.w() - 1 - pi;
                                if(pj+dj_p >= gradIntegralImage.h()) dj_p = gradIntegralImage.h() - 1 - pj;
                                if(pi-di_m < 0) di_m = pi;
                                if(pj-dj_m < 0) dj_m = pj;
                                
                                int ipij = pj * fpWeight.w() + pi;
                                fpWeight[ipij] = gradIntegralImage[fpWeight.w()*(pj+dj_p)+(pi+di_p)];
                                fpWeight[ipij] += gradIntegralImage[fpWeight.w()*(pj-dj_m)+(pi-di_m)];
                                fpWeight[ipij] -= gradIntegralImage[fpWeight.w()*(pj+dj_p)+(pi-di_m)];
                                fpWeight[ipij] -= gradIntegralImage[fpWeight.w()*(pj-dj_m)+(pi+di_p)];
                                
                                
                                //! We use |grad_u| instad of sqrt(|grad_u|^2 - sigma^2)
                                //! fpWeight[ipij] = sqrtf(fpWeight[ipij]);
                                
                            }
                        
                        
                        
                        
                        
                        //! Slow version
                        //fpWeight = fpGradient.patchMean(Diri_halfBlockSize);
                        
                        
                        //! avoid problems for points without any non zero weight
                        for (int kk = 0; kk < fpWeight.wh(); kk++) fpWeight[kk] = MAX(fpWeight[kk], 0.001f);
                        
                        
                    }
                    
                    
                    
                    //! accumulate transformed images
                    for(int kk = 0; kk < cfBursts.c(); kk++)
                    {
                        float *igray =  cfBursts.v(kk);
                        float *tgray  = idata[jj].v(kk);
                        
                        for (int ll=0; ll < idata[jj].wh(); ll++)
                            if (tgray[ll] > 0.0f) 
                                igray[ll] +=  fpWeight[ll] * tgray[ll];
                        
                    }
                    
                    
                    //! accumulate counter
                    float *tgray  =  idata[jj].v(0);
                    for (int ll=0; ll < idata[jj].wh(); ll++)
                        if (tgray[ll] > 0.0f)    fpCounter[ll] += fpWeight[ll];  					
                    
                    
                    
                }
                
                
                
                
                
                
                
                //! normalize burst image
                for(int kk = 0; kk < cfBursts.c(); kk++)
                {
                    float *igray =  cfBursts.v(kk);   
                    for (int ll=0; ll < cfBursts.wh(); ll++)  igray[ll] /= fpCounter[ll];
                    
                }
                
      
                
        
        
	}
	
    
    
    
    
    
    
    
    void histogram_specification_painting(float *in1, float *in2, float *out, int width1, int height1, int width2, int height2)
    {
        
        assert(in1 != NULL && in2 != NULL);
        
        if (width1 != width2 || height1 != height2)
        { printf("error :: fk_histogram_specification_painting :: images of different size"); exit(-1);}
        
        
        
        float *itmp1=new float[width1*height1];
        float *itmp2=new float[width2*height2];
        
        
        fpCopy(in1, itmp1, width1*height1);
        fpCopy(in2, itmp2, width2*height2);
        
        
        
        for (int ii=0; ii < width1 * height1; ii++)
            if (itmp1[ii] < 0.0f || itmp2[ii] < 0.0f)
            {
                itmp1[ii] = bg_here; itmp2[ii] = bg_here;
            }
        
        
        
        
        fk_histogram_specification(itmp1, itmp2, out,  width1,  height1,  width2,  height2);
        
        delete[] itmp1;
        delete[] itmp2;
        
        
        
    }
    
    
    
    
    
    void    combine_registered_bursts_min(cflimage *idata, int inImages, cflimage &cfCombined)
    {
        
        //! create image
        cfCombined.create(idata[0].w(), idata[0].h(), idata[0].c());
        
#ifdef _OPENMP
#pragma omp parallel for
#endif
        //! for each pixel
        for (int ix = 0; ix < cfCombined.w(); ix++)
            for (int iy=0; iy < cfCombined.h(); iy++)
            {
                
                
                float difMinim = fLarge;
                int ixy = iy * cfCombined.w() + ix;
                int index = -1;
                
                for (int jj = 0; jj < inImages; jj++) 
                    if ( idata[jj][ixy] > 0.0f)
                    {
                        
                        float dif = 0.0f;
                        for (int iC=0; iC < cfCombined.c(); iC++)   dif += fabsf(idata[jj][iC * cfCombined.wh() + ixy]);
                        
                        if (dif < difMinim) {difMinim=dif; index = jj;}
                    }  
                
                
                if (index != -1) 
                    for (int iC=0; iC < cfCombined.c(); iC++)   cfCombined[iC * cfCombined.wh() + ixy] = idata[index][iC * cfCombined.wh() + ixy];
                else for (int iC=0; iC < cfCombined.c(); iC++)   cfCombined[iC * cfCombined.wh() + ixy] = bg_here;
                
            }
    }
    
    
    
    
    void    combine_registered_bursts_medianRGB(cflimage *idata, int inImages, cflimage &cfCombined)
    {
        
        //! create image
        cfCombined.create(idata[0].w(), idata[0].h(), idata[0].c());
        
#ifdef _OPENMP
#pragma omp parallel for
#endif
        //! for each pixel
        for (int ix = 0; ix < cfCombined.w(); ix++)
            for (int iy=0; iy < cfCombined.h(); iy++)
            {
                
                
                float difMinim = fLarge;
                int index=-1;
                
                int ixy = iy * cfCombined.w() + ix;
                
                
                for (int jj = 0; jj < inImages; jj++) 
                    if ( idata[jj][ixy] > 0.0f)
                    {
                        
                        float dif = 0.0f;
                        float counter = 0.0f;
                        
                        for (int ii=0; ii < inImages; ii++)
                            if ( idata[ii][ixy] > 0.0f)
                            {
                                for (int iC=0; iC < cfCombined.c(); iC++)   dif += fabsf( idata[ii][iC * cfCombined.wh() + ixy] - idata[jj][iC * cfCombined.wh() + ixy]);
                                counter++;
                            }
                        
                        dif /= counter;
                        
                        if (dif < difMinim) {difMinim=dif; index = jj;}
                    }  
                
                if (index != -1)
                    for (int iC=0; iC < cfCombined.c(); iC++)   cfCombined[iC * cfCombined.wh() + ixy] = idata[index][iC * cfCombined.wh() + ixy];
                
            }
        
        
    }
    
    
    
    
    void  combine_registered_bursts_medianGrad(cflimage *idata, int inImages, cflimage &cfCombined)
    {
        
        
        
        //! create image
        //cfCombined.create(idata[0].w(), idata[0].h(), idata[0].c());
        cfCombined=idata[0];
        
        cflimage *xgrad = new cflimage[inImages];
        cflimage *ygrad = new cflimage[inImages];
        
        
        cflimage xgradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); xgradminim = 0.0f;
        cflimage ygradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); ygradminim = 0.0f;
        
        
        //! Compute gradients
#ifdef _OPENMP
#pragma omp parallel for
#endif
        for (int ii=0; ii < inImages; ii++)
        {
            libIIPStable::cflimage tmpOri;
            libIIPStable::cflimage tmpGrad = idata[ii].gradient(xgrad[ii], ygrad[ii], tmpOri,'f');
        }
        
        
        
#ifdef _OPENMP
#pragma omp parallel for
#endif
        //! Compute median of gradient vectors
        for (int ix = 1; ix < cfCombined.w() - 1; ix++)
            for (int iy = 1; iy < cfCombined.h() - 1; iy++)
            {
                
                
                float difMinim = fLarge;
                int index=-1;
                
                int ixy = iy * cfCombined.w() + ix;			
                int ixyp1 = iy * cfCombined.w() + ix + 1;			
                int ixym1 = iy * cfCombined.w() + ix - 1;			
                
                
                int ixypp1 = (iy + 1) * cfCombined.w() + ix ;			
                int ixymm1 = (iy - 1) * cfCombined.w() + ix ;	
                
                
                for (int jj = 0; jj < inImages; jj++) 
                    if ( idata[jj][ixy] >= 0.0f &&  idata[jj][ixyp1]>= 0.0f  &&  idata[jj][ixym1] >= 0.0f &&  idata[jj][ixypp1]>= 0.0f &&  idata[jj][ixymm1]>= 0.0f)
                    {
                        
                        float dif = 0.0f;
                        float counter = 0.0f;
                        
                        
                        float jjgradx = 0.0f, jjgrady = 0.0f;
                        for (int iC=0; iC < cfCombined.c(); iC++)
                        {
                            int iikk = iC * cfCombined.wh() + ixy;
                            jjgradx += xgrad[jj][iikk];
                            jjgrady += ygrad[jj][iikk];
                        }
                        
                        jjgradx /= (float) cfCombined.c();  jjgrady /= (float) cfCombined.c();
                        
                        for (int ii=0; ii < inImages; ii++)
                            if ( idata[ii][ixy] >= 0.0f &&  idata[ii][ixyp1]>= 0.0f  &&  idata[ii][ixym1] >= 0.0f &&  idata[ii][ixypp1]>= 0.0f &&  idata[ii][ixymm1]>= 0.0f)
                            {
                                
                                float iigradx = 0.0f, iigrady = 0.0f;
                                
                                
                                for (int iC=0; iC < cfCombined.c(); iC++)
                                {
                                    int iikk = iC * cfCombined.wh() + ixy;
                                    iigradx += 	 xgrad[ii][iikk];
                                    iigrady += 	 ygrad[ii][iikk];
                                }
                                
                                iigradx /= (float) cfCombined.c();  iigrady /= (float) cfCombined.c();
                                
                                
                                dif += fabsf(iigradx - jjgradx) +   fabsf(iigrady - jjgrady) ;
                                
                                counter++;
                            }
                        
                        dif /= counter;
                        
                        if (dif < difMinim) {difMinim=dif; index = jj;}
                    }  
                
                
                
                if (index != -1)
                {
                    for (int iC=0; iC < cfCombined.c(); iC++) 
                    {
                        xgradminim[iC * cfCombined.wh() + ixy] = xgrad[index][iC * cfCombined.wh() + ixy];
                        ygradminim[iC * cfCombined.wh() + ixy] = ygrad[index][iC * cfCombined.wh() + ixy];
                    }
                }
                
                
            }
        
        
        // give value to boundaries
        
        for (int ix = 0; ix < cfCombined.w(); ix++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + ix] = xgradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
                ygradminim[iC * cfCombined.wh() + ix] = ygradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
            }
            
        }	
        
        
        for (int iy = 0; iy < cfCombined.h(); iy++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
            }
            
        }	
        
        
        
        // Solve Poisson equation
        libIIPStable::solve_Poisson_fft(&cfCombined, &xgradminim, &ygradminim);
        
    }
    
    
    
    
    
    
    
    
    
    
    void  combine_registered_bursts_medianGrad_lm(cflimage *idata, int inImages, cflimage &cfCombined)
    {
        
        
        
        //! create image
        cfCombined=idata[0];
        
        
        //cflimage *xgrad = new cflimage[inImages];
        //cflimage *ygrad = new cflimage[inImages];
        
        
        cflimage xgradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); xgradminim = 0.0f;
        cflimage ygradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); ygradminim = 0.0f;
        
        
        //! Compute gradients
        //for (int ii=0; ii < inImages; ii++)
        //{
        //    libIIPStable::cflimage tmpOri;
        //    libIIPStable::cflimage tmpGrad = idata[ii].gradient(xgrad[ii], ygrad[ii], tmpOri,'f');
        //}
        
        
        
        //! Compute median of gradient vectors
        for (int ix = 1; ix < cfCombined.w() - 1; ix++)
            for (int iy = 1; iy < cfCombined.h() - 1; iy++)
            {
                
                
                float difMinim = fLarge;
                
                int index=-1;
               
                
                
                int ixy = iy * cfCombined.w() + ix;			
                int ixyp1 = iy * cfCombined.w() + ix + 1;			
                int ixym1 = iy * cfCombined.w() + ix - 1;			
                
                
                int ixypp1 = (iy + 1) * cfCombined.w() + ix ;			
                int ixymm1 = (iy - 1) * cfCombined.w() + ix ;	
                
                
                for (int jj = 0; jj < inImages; jj++) 
                    if ( idata[jj][ixy] >= 0.0f &&  idata[jj][ixyp1]>= 0.0f  &&  idata[jj][ixym1] >= 0.0f &&  idata[jj][ixypp1]>= 0.0f &&  idata[jj][ixymm1]>= 0.0f)
                    {
                        
                        float dif = 0.0f;
                        float counter = 0.0f;
                        
                        
                        float jjgradx = 0.0f, jjgrady = 0.0f;
                        
                        for (int iC=0; iC < cfCombined.c(); iC++)
                        {
                            float *ptr = idata[jj].v(iC);
                            
                            jjgradx +=  (ptr[ixy+1] - ptr[ixy]) ;
                            jjgrady +=  (ptr[ixy + idata[jj].w()] - ptr[ixy]);
                        }
                        

                        
                        jjgradx /= (float) cfCombined.c();  jjgrady /= (float) cfCombined.c();
                        
                        for (int ii=0; ii < inImages; ii++)
                            if ( idata[ii][ixy] >= 0.0f &&  idata[ii][ixyp1]>= 0.0f  &&  idata[ii][ixym1] >= 0.0f &&  idata[ii][ixypp1]>= 0.0f &&  idata[ii][ixymm1]>= 0.0f)
                            {
                                
                                float iigradx = 0.0f, iigrady = 0.0f;
                                
                                
                                for (int iC=0; iC < cfCombined.c(); iC++)
                                {
                                    float *ptr = idata[ii].v(iC);
                                                                 
                                    iigradx +=  (ptr[ixy+1] - ptr[ixy]) ;
                                    iigrady +=  (ptr[ixy + idata[ii].w()] - ptr[ixy]);
                                    
                                }
                                
                                
                                iigradx /= (float) cfCombined.c();  iigrady /= (float) cfCombined.c();
                                
                                
                                dif += fabsf(iigradx - jjgradx) +   fabsf(iigrady - jjgrady) ;
                                
                                counter++;
                                
                            }
                        
                        
                        dif /= counter;
                        
                        if (dif < difMinim) {difMinim=dif;  index = jj;}
                    }  
                
                
                
                if (index != -1)
                {			
                    for (int iC=0; iC < cfCombined.c(); iC++) 
                    {
                        float *ptr = idata[index].v(iC);
                        
                        xgradminim[iC * cfCombined.wh() + ixy] = (ptr[ixy+1] - ptr[ixy]) ;
                        ygradminim[iC * cfCombined.wh() + ixy] = (ptr[ixy + idata[index].w()] - ptr[ixy]) ;
                    }
                }
                
                
            }
        
        
        // give value to boundaries
        
        for (int ix = 0; ix < cfCombined.w(); ix++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + ix] = xgradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
                ygradminim[iC * cfCombined.wh() + ix] = ygradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
            }
            
        }	
        
        
        for (int iy = 0; iy < cfCombined.h(); iy++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
            }
            
        }	
        
        
        
        // Solve Poisson equation
        libIIPStable::solve_Poisson_fft(&cfCombined, &xgradminim, &ygradminim);
        
    }
    

    
    
    
    
    
    
    void  combine_registered_bursts_medianGrad_dif_res(cflimage *idata, int inImages, cflimage &cfCombined)
    {
        
        
        
        cfCombined = idata[0];
        
        
        libIIPStable::cflimage *xgrad = new libIIPStable::cflimage[inImages];
        libIIPStable::cflimage *ygrad = new libIIPStable::cflimage[inImages];
        
        
        libIIPStable::cflimage xgradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); xgradminim = 0.0f;
        libIIPStable::cflimage ygradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); ygradminim = 0.0f;
        
        
        // Compute gradients
        // CHECK ---> Gradient definition at boundary or near boundary.
        // Save derivatives ...
        for (int ii=0; ii < inImages; ii++)
        {
            libIIPStable::cflimage tmpOri;
            libIIPStable::cflimage tmpGrad = idata[ii].gradient(xgrad[ii], ygrad[ii], tmpOri,'f');
        }
        
        
        xgradminim = xgrad[0];
        ygradminim = ygrad[0];
        
        // Compute median of gradient vectors
        for (int ix = 1; ix < cfCombined.w() - 1; ix++)
            for (int iy = 1; iy < cfCombined.h() - 1; iy++)
            {
                
                
                float difMinim = fLarge;
                int index=-1;
                
                int ixy = iy * cfCombined.w() + ix;			
                int ixyp1 = iy * cfCombined.w() + ix + 1;			
                int ixym1 = iy * cfCombined.w() + ix - 1;			
                
                
                int ixypp1 = (iy + 1) * cfCombined.w() + ix ;			
                int ixymm1 = (iy - 1) * cfCombined.w() + ix ;	
                
                
                for (int jj = 1; jj < inImages; jj++) 
                    if ( idata[jj][ixy] >= 0.0f &&  idata[jj][ixyp1]>= 0.0f  &&  idata[jj][ixym1] >= 0.0f &&  idata[jj][ixypp1]>= 0.0f &&  idata[jj][ixymm1]>= 0.0f)
                    {
                        
                        float dif = 0.0f;
                        float counter = 0.0f;
                        
                        
                        float jjgradx = 0.0f, jjgrady = 0.0f;
                        for (int iC=0; iC < cfCombined.c(); iC++)
                        {
                            int iikk = iC * cfCombined.wh() + ixy;
                            jjgradx += xgrad[jj][iikk];
                            jjgrady += ygrad[jj][iikk];
                        }
                        
                        jjgradx /= (float) cfCombined.c();  jjgrady /= (float) cfCombined.c();
                        
                        for (int ii=1; ii < inImages; ii++)
                            if ( idata[ii][ixy] >= 0.0f &&  idata[ii][ixyp1]>= 0.0f  &&  idata[ii][ixym1] >= 0.0f &&  idata[ii][ixypp1]>= 0.0f &&  idata[ii][ixymm1]>= 0.0f)
                            {
                                
                                float iigradx = 0.0f, iigrady = 0.0f;
                                
                                
                                for (int iC=0; iC < cfCombined.c(); iC++)
                                {
                                    int iikk = iC * cfCombined.wh() + ixy;
                                    iigradx += 	 xgrad[ii][iikk];
                                    iigrady += 	 ygrad[ii][iikk];
                                }
                                
                                iigradx /= (float) cfCombined.c();  iigrady /= (float) cfCombined.c();
                                
                                
                                dif += fabsf(iigradx - jjgradx) +   fabsf(iigrady - jjgrady) ;
                                
                                counter++;
                            }
                        
                        dif /= counter;
                        
                        if (dif < difMinim) {difMinim=dif; index = jj;}
                    }  
                
                
                
                if (index != -1)
                {			
                    for (int iC=0; iC < cfCombined.c(); iC++) 
                    {
                        xgradminim[iC * cfCombined.wh() + ixy] = xgrad[index][iC * cfCombined.wh() + ixy];
                        ygradminim[iC * cfCombined.wh() + ixy] = ygrad[index][iC * cfCombined.wh() + ixy];
                    }
                }
                
                
            }
        
        
        // give value to boundaries
        
        for (int ix = 0; ix < cfCombined.w(); ix++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + ix] = xgradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
                ygradminim[iC * cfCombined.wh() + ix] = ygradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
            }
            
        }	
        
        
        for (int iy = 0; iy < cfCombined.h(); iy++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
            }
            
        }	
        
        
        
        // Solve Poisson equation
        libIIPStable::solve_Poisson_fft(&cfCombined, &xgradminim, &ygradminim);
        
    }
    
    
    
    
    
    void  combine_registered_bursts_medianGrad_dif_res_lm(cflimage *idata, int inImages, cflimage &cfCombined)
    {
        
        
        
        //! create image
        cfCombined=idata[0];
        
        
        //cflimage *xgrad = new cflimage[inImages];
        //cflimage *ygrad = new cflimage[inImages];
        
        
        cflimage xgradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); xgradminim = 0.0f;
        cflimage ygradminim(cfCombined.w(), cfCombined.h(), cfCombined.c()); ygradminim = 0.0f;
        
        
        //! Compute gradients
        //for (int ii=0; ii < inImages; ii++)
        //{
        //    libIIPStable::cflimage tmpOri;
        //    libIIPStable::cflimage tmpGrad = idata[ii].gradient(xgrad[ii], ygrad[ii], tmpOri,'f');
        //}
        
        
        
     
        
        //! Compute median of gradient vectors
        for (int ix = 1; ix < cfCombined.w() - 1; ix++)
            for (int iy = 1; iy < cfCombined.h() - 1; iy++)
            {
                
          
                
                
                float difMinim = fLarge;
                
                int index=-1;
                
                
                
                int ixy = iy * cfCombined.w() + ix;			
                int ixyp1 = iy * cfCombined.w() + ix + 1;			
                int ixym1 = iy * cfCombined.w() + ix - 1;			
                
                
                int ixypp1 = (iy + 1) * cfCombined.w() + ix ;			
                int ixymm1 = (iy - 1) * cfCombined.w() + ix ;	
                
                
                
                
                //! Initialize gradient
                for (int iC=0; iC < cfCombined.c(); iC++) 
                {
                    float *ptr = idata[0].v(iC);
                    
                    xgradminim[iC * cfCombined.wh() + ixy] = (ptr[ixy+1] - ptr[ixy]) ;
                    ygradminim[iC * cfCombined.wh() + ixy] = (ptr[ixy + idata[0].w()] - ptr[ixy]) ;
                }
                

                
                
                
                
                
                
                for (int jj = 1; jj < inImages; jj++) 
                    if ( idata[jj][ixy] >= 0.0f &&  idata[jj][ixyp1]>= 0.0f  &&  idata[jj][ixym1] >= 0.0f &&  idata[jj][ixypp1]>= 0.0f &&  idata[jj][ixymm1]>= 0.0f)
                    {
                        
                        float dif = 0.0f;
                        float counter = 0.0f;
                        
                        
                                               
                        
                        float jjgradx = 0.0f, jjgrady = 0.0f;
                        
                        for (int iC=0; iC < cfCombined.c(); iC++)
                        {
                            float *ptr = idata[jj].v(iC);
                            
                            jjgradx +=  (ptr[ixy+1] - ptr[ixy]) ;
                            jjgrady +=  (ptr[ixy + idata[jj].w()] - ptr[ixy]);
                        }
                        
                        
                        
                        jjgradx /= (float) cfCombined.c();  jjgrady /= (float) cfCombined.c();
                        
                        for (int ii=1; ii < inImages; ii++)
                            if ( idata[ii][ixy] >= 0.0f &&  idata[ii][ixyp1]>= 0.0f  &&  idata[ii][ixym1] >= 0.0f &&  idata[ii][ixypp1]>= 0.0f &&  idata[ii][ixymm1]>= 0.0f)
                            {
                                
                                float iigradx = 0.0f, iigrady = 0.0f;
                                
                                
                                for (int iC=0; iC < cfCombined.c(); iC++)
                                {
                                    float *ptr = idata[ii].v(iC);
                                    
                                    iigradx +=  (ptr[ixy+1] - ptr[ixy]) ;
                                    iigrady +=  (ptr[ixy + idata[ii].w()] - ptr[ixy]);
                                    
                                }
                                
                                
                                iigradx /= (float) cfCombined.c();  iigrady /= (float) cfCombined.c();
                                
                                
                                dif += fabsf(iigradx - jjgradx) +   fabsf(iigrady - jjgrady) ;
                                
                                counter++;
                                
                            }
                        
                        
                        dif /= counter;
                        
                        if (dif < difMinim) {difMinim=dif;  index = jj;}
                    }  
                
                
                
                if (index != -1)
                {			
                    for (int iC=0; iC < cfCombined.c(); iC++) 
                    {
                        float *ptr = idata[index].v(iC);
                        
                        xgradminim[iC * cfCombined.wh() + ixy] = (ptr[ixy+1] - ptr[ixy]) ;
                        ygradminim[iC * cfCombined.wh() + ixy] = (ptr[ixy + idata[index].w()] - ptr[ixy]) ;
                    }
                }
                
                
            }
        
        
        // give value to boundaries
        
        for (int ix = 0; ix < cfCombined.w(); ix++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + ix] = xgradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =xgradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
                ygradminim[iC * cfCombined.wh() + ix] = ygradminim[iC * cfCombined.wh() +  cfCombined.w() + ix] ;
                ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 1) * cfCombined.w() +  ix] =ygradminim[iC * cfCombined.wh() + (cfCombined.h() - 2) * cfCombined.w() + ix];
                
                
            }
            
        }	
        
        
        for (int iy = 0; iy < cfCombined.h(); iy++)
        {
            for (int iC=0; iC < cfCombined.c(); iC++) 
            {
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = xgradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() ] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + 1] ;
                ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 1] = ygradminim[iC * cfCombined.wh() + iy * cfCombined.w() + cfCombined.w() - 2] ;
                
            }
            
        }	
        
        
        
        // Solve Poisson equation
        libIIPStable::solve_Poisson_fft(&cfCombined, &xgradminim, &ygradminim);
        
    }
    
    
    
    
}












/*
 
 namespace libIIPStable
 {
 
 
 
 
 
 
 
 
 
 
 
 
 
 void    combine_registered_bursts_medianRGB_dif_res(libIIPStable::cflimage *idata, int inImages, libIIPStable::cflimage &cfCombined)
 {
 
 cfCombined = idata[0];
 
 
 for (int ix = 0; ix < cfCombined.w(); ix++)
 for (int iy=0; iy < cfCombined.h(); iy++)
 {
 
 
 float difMinim = fLarge;
 int index=-1;
 
 int ixy = iy * cfCombined.w() + ix;
 
 
 for (int jj = 1; jj < inImages; jj++) 
 if ( idata[jj][ixy] > 0.0f)
 {
 
 float dif = 0.0f;
 float counter = 0.0f;
 
 for (int ii = 1; ii < inImages; ii++)
 if ( idata[ii][ixy] > 0.0f)
 {
 for (int iC=0; iC < cfCombined.c(); iC++)   dif += fabsf( idata[ii][iC * cfCombined.wh() + ixy] - idata[jj][iC * cfCombined.wh() + ixy]);
 counter++;
 }
 
 dif /= counter;
 
 if (dif < difMinim) {difMinim=dif; index = jj;}
 }  
 
 
 if (index != -1)
 for (int iC=0; iC < cfCombined.c(); iC++)   cfCombined[iC * cfCombined.wh() + ixy] = idata[index][iC * cfCombined.wh() + ixy];
 
 }
 
 
 }
 
 
 
 
 
 
 
 
 */




/*
 
 void build_bursts(cflimage *idata, cflimage *odata, int inImages, int &inBursts, int *fiIndexos, int *fiNumber, int iVerbose)
 {
 
 
 //--//   Parameters  // -- //
 for (int ii=0; ii < inImages; ii++) fiIndexos[ii] = 0;
 fiIndexos[0] = 0;
 fiNumber[0] = 1;
 inBursts = 1;
 
 
 
 //! sift parameters
 siftPar parameters;		
 default_sift_parameters(parameters);
 
 parameters.OctaveMax = 5;
 parameters.PeakThresh = 2.0f;
 
 //parameters.MatchXradius = 0.25 * MAX(idata[0].w(), idata[0].h());
 //parameters.MatchYradius = 0.25 * MAX(idata[0].w(), idata[0].h());
 
 
 //! ransac parameters
 int inRansacIter = 1000;
 float fRansacTol = 1.50f;
 
 //!  minimum number of matches to compute an homography
 int inAccordedHom = 10;
 
 
 //! maximum distance par image corner allowed under homography transformation to be considered as a burst
 float fBurstHomTol = 0.10 * MAX(idata[0].w(), idata[0].h());
 
 
 //! print parameters
 if (iVerbose)
 {	
 std::cout << "\t... Ransac.\t\tIter: " << inRansacIter << "   Tol: " << fRansacTol << std::endl;
 std::cout << "\t... Homography.\t\tinMin:	" << inAccordedHom  << "   Tolerance same burst:  " << fBurstHomTol <<  std::endl << std::endl;
 }
 
 
 //! memory for homography
 laMatrix H(3,3), Hinv(3,3);
 
 
 //! memory for svd decomposition
 laVector Dsvd(2);
 laMatrix  Usvd(2,2), Vsvd(2,2), Asvd(2,2);
 
 
 //--// Dealing with first image //--//
 
 cflimage init = idata[0];
 flimage tmp_gray_init = init.getGray();     // convert it to gray
 
 
 keypointslist keys0;                        // keypoints of initial image
 compute_sift_keypoints(tmp_gray_init.v(), keys0, init.w(), init.h(), parameters);
 
 tmp_gray_init.erase();                              // erase temporal gray image
 
 
 //! Fill output first image
 odata[0] = idata[0];
 
 
 
 
 for(int iiIndex = 1; iiIndex < inImages; iiIndex++)
 {
 
 
 //--// Dealing with next image //--//		
 
 cflimage image=idata[iiIndex];
 flimage tmp_gray_image = image.getGray();						// temporal tmp_gray_image
 
 keypointslist keys;												// keypoints of current image
 
 compute_sift_keypoints(tmp_gray_image.v() ,keys, image.w(),image.h(),parameters);
 
 tmp_gray_image.erase();											// erase tmp_gray_image
 
 matchingslist matchings;										// apply sift
 compute_sift_matches(keys0,keys,matchings,parameters);
 
 
 //--//   RANSAC   //--//
 
 int n = matchings.size();										
 int naccorded=0;				
 if (n > inAccordedHom)											// if enough matches
 {
 
 float *x0 = new float[n]; float *y0 = new float[n]; 
 float *x1 = new float[n]; float *y1 = new float[n];
 
 matchingslist::iterator ptr = matchings.begin();			// store matches in float pointers
 for(int jj=0; jj < n; jj++, ptr++) 
 {
 x0[jj] = ptr->first.x;  y0[jj] = ptr->first.y;
 x1[jj] = ptr->second.x; y1[jj] = ptr->second.y;
 }
 
 
 int *accorded = new int[n];									// compute homography by applying Ransac
 
 naccorded = compute_ransac_planar_homography_n_points(x0,y0,x1,y1,n, inRansacIter, fRansacTol, H);
 
 
 
 delete[] x0; delete[] y0;
 delete[] x1; delete[] y1;
 delete[] accorded;
 
 } 
 
 
 
 
 //--//   applying homography   //--//
 if (iVerbose) std::cout << "\t(" << fiIndexos[inBursts-1] << " , " << iiIndex << " ):  " << n << " , " << naccorded  ; 
 
 
 // normalize homography
 double h1,h2,dMaxDist,dTilt;
 
 if (naccorded > inAccordedHom)											// if enough matches
 {
 
 if (H[2][2] != 0.0f)										
 for(int i=0; i < 3; i++) 
 for(int j=0; j < 3; j++) H[i][j] /=  H[2][2];
 
 
 // h1, h2 to check if it is an affinity
 h1 = H[2][0];
 h2 = H[2][1];
 
 // distance of mapped corners
 dMaxDist = max_dist_vertex(H, init.w(), init.h());   
 
 // tilt
 Asvd[0][0]=H[0][0];
 Asvd[0][1]=H[0][1];
 Asvd[1][0]=H[1][0];
 Asvd[1][1]=H[1][1];
 
 compute_svd(Asvd, Usvd, Vsvd, Dsvd);
 
 
 dTilt = MAX(Dsvd[0],Dsvd[1]);
 dTilt /= MIN(Dsvd[0], Dsvd[1]);
 }
 
 
 
 if (naccorded > inAccordedHom && ((dMaxDist < fBurstHomTol && fabsf(h1) < 0.0001 & fabsf(h2) < 0.0001 && dTilt < 1.05)) )
 {
 if (iVerbose) std::cout << "  --> accepted (dMaxDist, h1, h2, Tilt) :  ("  << dMaxDist << " , " << h1 << " , " << h2 << " , " << dTilt << " )" << std::endl;
 
 
 luinv(H, Hinv);									// inverse corresponding homograpy
 
 
 //! transformed image has same size than initial
 odata[iiIndex].create(init.w(), init.h(), init.c());
 odata[iiIndex] = BagValueAfterRegistration;
 
 
 //! use mask to avoid boundary effects and apply erosion
 flimage tmpMaskAux(image.w(), image.h());			
 flimage tmpMask(init.w(), init.h());
 
 tmpMaskAux = 255.0f;
 tmpMask = BagValueAfterRegistration;
 
 apply_planar_homography(tmpMaskAux.v(),tmpMaskAux.w(),tmpMaskAux.h(), Hinv, BagValueAfterRegistration, HOMOGRAPHY_RESAMPLING_ORDER,tmpMask.v(),0.0,0.0,tmpMask.w(),tmpMask.h());				
 
 tmpMaskAux = tmpMask;
 fiPatchMin(tmpMaskAux.v(), tmpMask.v(), 3.0, tmpMask.w(), tmpMask.h());
 
 
 
 //! Transform images
 for (int ii=0; ii < image.c(); ii++)
 {
 apply_planar_homography(image.v(ii),image.w(),image.h(), Hinv, BagValueAfterRegistration, HOMOGRAPHY_RESAMPLING_ORDER, odata[iiIndex].v(ii),0.0,0.0,odata[iiIndex].w(),odata[iiIndex].h());
 
 for (int ijk=0; ijk < odata[iiIndex].wh(); ijk++) 
 {
 if (tmpMask[ijk] < 0.0f) odata[iiIndex][ii * odata[iiIndex].wh() + ijk] = BagValueAfterRegistration;
 
 }
 
 }
 
 
 //! update number of images in burst
 fiNumber[inBursts-1]++;
 
 
 }  else  //--//   It's not the same burst  //--//
 {
 if (iVerbose) std::cout << "  --> rejected (dMaxDist, h1, h2, Tilt) :  ("  << dMaxDist << " , " <<  h1 << " , " << h2 << " , " << dTilt << " )" << std::endl;
 
 fiIndexos[inBursts] = iiIndex;
 fiNumber[inBursts] = 1;
 
 init = idata[iiIndex];
 odata[iiIndex] = idata[iiIndex];
 
 keys0 = keys;
 
 inBursts++;
 }
 
 
 }
 
 }
 */



