#include "libAsiftIPOL.h"



namespace libSiftIPOLCODE 
{
    
    
    
    
#define ABS(x)    (((x) > 0) ? (x) : (-(x)))
    
    
    static float InitSigma_aa = 1.6;
    
#define round(x) ((x)>=0?(long)((x)+0.5):(long)((x)-0.5))
    
    const float GaussTruncate1 = 4.0;
    
 
    
    
    bool loadMatch(const char* nameFile, std::vector<Match>& match) {
        match.clear();
        std::ifstream f(nameFile);
        while( f.good() ) {
            std::string str;
            std::getline(f, str);
            if( f.good() ) {
                std::istringstream s(str);
                Match m;
                s >> m.x1 >> m.y1 >> m.x2 >> m.y2;
                if(!s.fail() )
                    match.push_back(m);
            }
        }
        return !match.empty();
    }
    
    bool saveMatch(const char* nameFile, const std::vector<Match>& match) {
        std::ofstream f(nameFile);
        if( f.is_open() ) {
            std::vector<Match>::const_iterator it = match.begin();
            for(; it != match.end(); ++it)
                f << it->x1 << " " << it->y1 << " "
                << it->x2 << " " << it->y2 << std::endl;
        }
        return f.is_open();
    }

    
    
    
    
    
    double initcausal(double *c,int n,double z)
    {
        double zk,z2k,iz,sum;
        int k;
        
        zk = z; iz = 1./z;
        z2k = pow(z,(double)n-1.);
        sum = c[0] + z2k * c[n-1];
        z2k = z2k*z2k*iz;
        for (k=1;k<=n-2;k++) {
            sum += (zk+z2k)*c[k];
            zk *= z;
            z2k *= iz;
        }
        return (sum/(1.-zk*zk));
    }
    
    double initanticausal(double *c,int n,double z)
    {
        return((z/(z*z-1.))*(z*c[n-2]+c[n-1]));
    }
    
    
    void invspline1D(double *c,int size,double *z,int npoles)
    {
        double lambda;
        int n,k;
        
        /* normalization */
        for (k=npoles,lambda=1.;k--;) lambda *= (1.-z[k])*(1.-1./z[k]);
        for (n=size;n--;) c[n] *= lambda;
        
        /*----- Loop on poles -----*/
        for (k=0;k<npoles;k++) {
            
            /* forward recursion */
            c[0] = initcausal(c,size,z[k]);
            for (n=1;n<size;n++) 
                c[n] += z[k]*c[n-1];
            
            /* backwards recursion */
            c[size-1] = initanticausal(c,size,z[k]);
            for (n=size-1;n--;) 
                c[n] = z[k]*(c[n+1]-c[n]);
            
        }
    }
    
    
    /*------------------------------ MAIN MODULE ------------------------------*/
    
    // void finvspline(float *in,int order,float *out, int width, int height)
    
    // Guoshen Yu, 2010.09.21, Windows version
    void finvspline(std::vector<float> &in,int order,std::vector<float>& out, int width, int height)
    // void finvspline(float *in,int order,float *out, int width, int height)
    {
        double *c,*d,z[5];
        int npoles,nx,ny,x,y;
        
        ny = height; nx = width;
        
        /* initialize poles of associated z-filter */
        switch (order) 
        {
            case 2: z[0]=-0.17157288;  /* sqrt(8)-3 */
                break;
                
            case 3: z[0]=-0.26794919;  /* sqrt(3)-2 */ 
                break;
                
            case 4: z[0]=-0.361341; z[1]=-0.0137254;
                break;
                
            case 5: z[0]=-0.430575; z[1]=-0.0430963;
                break;
                
            case 6: z[0]=-0.488295; z[1]=-0.0816793; z[2]=-0.00141415;
                break;
                
            case 7: z[0]=-0.53528; z[1]=-0.122555; z[2]=-0.00914869;
                break;
                
            case 8: z[0]=-0.574687; z[1]=-0.163035; z[2]=-0.0236323; z[3]=-0.000153821;
                break;
                
            case 9: z[0]=-0.607997; z[1]=-0.201751; z[2]=-0.0432226; z[3]=-0.00212131;
                break;
                
            case 10: z[0]=-0.636551; z[1]=-0.238183; z[2]=-0.065727; z[3]=-0.00752819;
                z[4]=-0.0000169828;
                break;
                
            case 11: z[0]=-0.661266; z[1]=-0.27218; z[2]=-0.0897596; z[3]=-0.0166696; 
                z[4]=-0.000510558;
                break;
                
            default:
                printf("finvspline: order should be in 2..11.\n");
                exit(-1);
        }
        
        npoles = order/2;
        
        /* initialize double array containing image */
        c = (double *)malloc(nx*ny*sizeof(double));
        d = (double *)malloc(nx*ny*sizeof(double));
        for (x=nx*ny;x--;) 
            c[x] = (double)in[x];
        
        /* apply filter on lines */
        for (y=0;y<ny;y++) 
            invspline1D(c+y*nx,nx,z,npoles);
        
        /* transpose */
        for (x=0;x<nx;x++)
            for (y=0;y<ny;y++) 
                d[x*ny+y] = c[y*nx+x];
        
        /* apply filter on columns */
        for (x=0;x<nx;x++) 
            invspline1D(d+x*ny,ny,z,npoles);
        
        /* transpose directy into image */
        for (x=0;x<nx;x++)
            for (y=0;y<ny;y++) 
                out[y*nx+x] = (float)(d[x*ny+y]);
        
        /* free array */
        free(d);
        free(c);
    }
    
    
    
    /* extract image value (even outside image domain) */
    //float v(float *in,int x,int y,float bg, int width, int height)
    // Guoshen Yu, 2010.09.21, Windows version
    float v(std::vector<float>& in,int x,int y,float bg, int width, int height)
    // float v(float *in, int x,int y,float bg, int width, int height)
    {
        if (x<0 || x>=width || y<0 || y>=height)
            return(bg); else return(in[y*width+x]);
    }
    
    
    /* c[] = values of interpolation function at ...,t-2,t-1,t,t+1,... */
    
    /* coefficients for cubic interpolant (Keys' function) */
    void keys(float *c,float t,float a)
    {
        float t2,at;
        
        t2 = t*t;
        at = a*t;
        c[0] = a*t2*(1.0-t);
        c[1] = (2.0*a+3.0 - (a+2.0)*t)*t2 - at;
        c[2] = ((a+2.0)*t - a-3.0)*t2 + 1.0;
        c[3] = a*(t-2.0)*t2 + at;
    }
    
    /* coefficients for cubic spline */
    void spline3(float *c,float t)
    {
        float tmp;
        
        tmp = 1.-t;
        c[0] = 0.1666666666*t*t*t;
        c[1] = 0.6666666666-0.5*tmp*tmp*(1.+t);
        c[2] = 0.6666666666-0.5*t*t*(2.-t);
        c[3] = 0.1666666666*tmp*tmp*tmp;
    }
    
    /* pre-computation for spline of order >3 */
    void init_splinen(float *a,int n)
    {
        int k;
        
        a[0] = 1.;
        for (k=2;k<=n;k++) a[0]/=(float)k;
        for (k=1;k<=n+1;k++)
            a[k] = - a[k-1] *(float)(n+2-k)/(float)k;
    }
    
    /* fast integral power function */
    float ipow(float x,int n)
    {
        float res;
        
        for (res=1.;n;n>>=1) {
            if (n&1) res*=x;
            x*=x;
        }
        return(res);
    }
    
    /* coefficients for spline of order >3 */
    void splinen(float *c,float t,float *a,int n)
    {
        int i,k;
        float xn;
        
        memset((void *)c,0,(n+1)*sizeof(float));
        for (k=0;k<=n+1;k++) { 
            xn = ipow(t+(float)k,n);
            for (i=k;i<=n;i++) 
                c[i] += a[i-k]*xn;
        }
    }
    
    
    
    
    
    void bound(int x, int y, float ca, float sa, int *xmin, int *xmax, int *ymin, int *ymax);
    
    
    /* NB : calling this module with out=in is nonsense */
    
    /* void frot(in,out,a,b,k_flag)
     Fimage in,out;
     float  *a,*b;
     char *k_flag; */
    
    void frot(std::vector<float>& in, std::vector<float>& out, int nx, int ny, int *nx_out, int *ny_out, float *a, float *b, char *k_flag)
    //void frot(float *in, float *out, int nx, int ny, int *nx_out, int *ny_out, float *a, float *b, char *k_flag)
    {
        /* int    nx,ny,x,y,x1,y1,adr; */
        int    x,y,x1,y1,adr; 
        float  ca,sa,xp,yp,a11,a12,a21,a22,ux,uy,xtrans,ytrans;
        int    tx1,ty1,tx2,ty2,xmin,xmax,ymin,ymax,sx,sy;
        
        /*    nx = in->ncol;
         ny = in->nrow; */
        
        ca = (float)cos((double)(*a)*M_PI/180.0);
        sa = (float)sin((double)(*a)*M_PI/180.0);
        
        
        
        /********** Compute new image location **********/
        if (k_flag) {
            /* crop image and fix center */
            xmin = ymin = 0;
            xmax = nx-1;
            ymax = ny-1;
            xtrans = 0.5*( (float)(nx-1)*(1.0-ca)+(float)(ny-1)*sa );
            ytrans = 0.5*( (float)(ny-1)*(1.0-ca)-(float)(nx-1)*sa );
        } else {
            /* extend image size to include the whole input image */
            xmin = xmax = ymin = ymax = 0;
            bound(nx-1,0,ca,sa,&xmin,&xmax,&ymin,&ymax);
            bound(0,ny-1,ca,sa,&xmin,&xmax,&ymin,&ymax);
            bound(nx-1,ny-1,ca,sa,&xmin,&xmax,&ymin,&ymax);
            xtrans = ytrans = 0.0;
        }
        sx = xmax-xmin+1;
        sy = ymax-ymin+1;
        
        /* out = mw_change_fimage(out,sy,sx);
         if (!out) mwerror(FATAL,1,"not enough memory\n"); */
        
        *nx_out = sx;
        *ny_out = sy;
        
        //	printf("Hello sx=%d, sy=%d\n", sx, sy);
        
        //		out = new float[sy*sx];
        out = std::vector<float>(sx*sy);
        
        /********** Rotate image **********/
        for (x=xmin;x<=xmax;x++) 
            for (y=ymin;y<=ymax;y++) {
                xp = ca*(float)x-sa*(float)y + xtrans;
                yp = sa*(float)x+ca*(float)y + ytrans;
                x1 = (int)floor(xp); 
                y1 = (int)floor(yp); 
                ux = xp-(float)x1;
                uy = yp-(float)y1;
                adr = y1*nx+x1;
                tx1 = (x1>=0 && x1<nx);
                tx2 = (x1+1>=0 && x1+1<nx);
                ty1 = (y1>=0 && y1<ny);
                ty2 = (y1+1>=0 && y1+1<ny);
                
                /*	a11 = (tx1 && ty1? in->gray[adr]:*b);
                 a12 = (tx1 && ty2? in->gray[adr+nx]:*b);
                 a21 = (tx2 && ty1? in->gray[adr+1]:*b);
                 a22 = (tx2 && ty2? in->gray[adr+nx+1]:*b); */
                
                a11 = (tx1 && ty1? in[adr]:*b);
                a12 = (tx1 && ty2? in[adr+nx]:*b);
                a21 = (tx2 && ty1? in[adr+1]:*b);
                a22 = (tx2 && ty2? in[adr+nx+1]:*b); 
                
                
                /*	out->gray[(y-ymin)*sx+x-xmin] = 
                 (1.0-uy)*((1.0-ux)*a11+ux*a21)+uy*((1.0-ux)*a12+ux*a22);*/
                
                out[(y-ymin)*sx+x-xmin] = 
                (1.0-uy)*((1.0-ux)*a11+ux*a21)+uy*((1.0-ux)*a12+ux*a22);
                
            }
    }
    
    
    void bound(int x, int y, float ca, float sa, int *xmin, int *xmax, int *ymin, int *ymax)
    /* int x,y;
     float ca,sa;
     int *xmin,*xmax,*ymin,*ymax;*/
    {   
        int rx,ry;
        
        rx = (int)floor(ca*(float)x+sa*(float)y);
        ry = (int)floor(-sa*(float)x+ca*(float)y);
        if (rx<*xmin) *xmin=rx; if (rx>*xmax) *xmax=rx;
        if (ry<*ymin) *ymin=ry; if (ry>*ymax) *ymax=ry;
    }

    
    
    
    //void fproj(float *in, float *out, int nx, int ny, int *sx, int *sy, float *bg, int *o, float *p, char *i, float X1, float Y1, float X2, float Y2, float X3, float Y3, float *x4, float *y4)
    void fproj(std::vector<float>& in, std::vector<float>& out, int nx, int ny, int *sx, int *sy, float *bg, int *o, float *p, char *i, float X1, float Y1, float X2, float Y2, float X3, float Y3, float *x4, float *y4)
    /*     Fimage in,out;
     int    *sx,*sy,*o;
     char   *i;
     float  *bg,*p,X1,Y1,X2,Y2,X3,Y3,*x4,*y4; */
    {
        /*  int    n1,n2,nx,ny,x,y,xi,yi,adr,dx,dy;*/
        int    n1,n2,x,y,xi,yi,adr,dx,dy;
        float  res,xx,yy,xp,yp,ux,uy,a,b,d,fx,fy,x12,x13,y12,y13;
        float  cx[12],cy[12],ak[13];
        /* Fimage ref,coeffs; */
        //  float *ref, *coeffs;
	std::vector<float> ref, coeffs;
        
        
        /* CHECK ORDER */
        if (*o!=0 && *o!=1 && *o!=-3 && 
            *o!=3 && *o!=5 && *o!=7 && *o!=9 && *o!=11)
        /*  mwerror(FATAL,1,"unrecognized interpolation order.\n"); */
        {	
            printf("unrecognized interpolation order.\n");
            exit(-1);
        }
        
        /* ALLOCATE NEW IMAGE */
        /*  nx = in->ncol; ny = in->nrow; */
        /*  out = mw_change_fimage(out,*sy,*sx); 
         if (!out) mwerror(FATAL,1,"not enough memory\n"); */
        
        
        if (*o>=3) {
            /*    coeffs = mw_new_fimage();
             finvspline(in,*o,coeffs); */
            
            //	  coeffs = new float[nx*ny];
            
            coeffs = std::vector<float>(nx*ny);
            
            finvspline(in,*o,coeffs,nx,ny);
            
            ref = coeffs;
            if (*o>3) init_splinen(ak,*o);
        } else {
            //    coeffs = NULL;
            ref = in;
        }
        
        
        /* COMPUTE NEW BASIS */
        if (i) {
            x12 = (X2-X1)/(float)nx;
            y12 = (Y2-Y1)/(float)nx;
            x13 = (X3-X1)/(float)ny;
            y13 = (Y3-Y1)/(float)ny;
        } else {
            x12 = (X2-X1)/(float)(*sx);
            y12 = (Y2-Y1)/(float)(*sx);
            x13 = (X3-X1)/(float)(*sy);
            y13 = (Y3-Y1)/(float)(*sy);
        }
        
        
        
        if (y4) { 
            xx=((*x4-X1)*(Y3-Y1)-(*y4-Y1)*(X3-X1))/((X2-X1)*(Y3-Y1)-(Y2-Y1)*(X3-X1));
            yy=((*x4-X1)*(Y2-Y1)-(*y4-Y1)*(X2-X1))/((X3-X1)*(Y2-Y1)-(Y3-Y1)*(X2-X1));
            a = (yy-1.0)/(1.0-xx-yy);
            b = (xx-1.0)/(1.0-xx-yy);
        } 
        else 
        {
            a=b=0.0;
        }
        
        
        
        
        /********** MAIN LOOP **********/
        
        for (x=0;x<*sx;x++) 
            for (y=0;y<*sy;y++) {
                
                /* COMPUTE LOCATION IN INPUT IMAGE */
                if (i) {
                    xx = 0.5+(((float)x-X1)*y13-((float)y-Y1)*x13)/(x12*y13-y12*x13);
                    yy = 0.5-(((float)x-X1)*y12-((float)y-Y1)*x12)/(x12*y13-y12*x13);
                    d = 1.0-(a/(a+1.0))*xx/(float)nx-(b/(b+1.0))*yy/(float)ny;
                    xp = xx/((a+1.0)*d);
                    yp = yy/((b+1.0)*d);
                } else {
                    fx = (float)x + 0.5;
                    fy = (float)y + 0.5;
                    d = a*fx/(float)(*sx)+b*fy/(float)(*sy)+1.0;
                    xx = (a+1.0)*fx/d;
                    yy = (b+1.0)*fy/d;
                    xp = X1 + xx*x12 + yy*x13;
                    yp = Y1 + xx*y12 + yy*y13;
                }
                
                
                /* INTERPOLATION */
                
                if (*o==0) { 
                    
                    /* zero order interpolation (pixel replication) */
                    xi = (int)floor((double)xp); 
                    yi = (int)floor((double)yp);
                    /*	if (xi<0 || xi>=in->ncol || yi<0 || yi>=in->nrow)*/
                    if (xi<0 || xi>=nx || yi<0 || yi>=ny)
                        res = *bg; 
                    else 
                    /* res = in->gray[yi*in->ncol+xi]; */
                        res = in[yi*nx+xi];
                } else { 
                    
                    /* higher order interpolations */
                    if (xp<0. || xp>(float)nx || yp<0. || yp>(float)ny) res=*bg; 
                    else {
                        xp -= 0.5; yp -= 0.5;
                        xi = (int)floor((double)xp); 
                        yi = (int)floor((double)yp);
                        ux = xp-(float)xi;
                        uy = yp-(float)yi;
                        switch (*o) 
                        {
                            case 1: /* first order interpolation (bilinear) */
                                n2 = 1;
                                cx[0]=ux;	cx[1]=1.-ux;
                                cy[0]=uy; cy[1]=1.-uy;
                                break;
                                
                            case -3: /* third order interpolation (bicubic Keys' function) */
                                n2 = 2;
                                keys(cx,ux,*p);
                                keys(cy,uy,*p);
                                break;
                                
                            case 3: /* spline of order 3 */
                                n2 = 2;
                                spline3(cx,ux);
                                spline3(cy,uy);
                                break;
                                
                            default: /* spline of order >3 */
                                n2 = (1+*o)/2;
                                splinen(cx,ux,ak,*o);
                                splinen(cy,uy,ak,*o);
                                break;
                        }
                        
                        res = 0.; n1 = 1-n2;
                        /* this test saves computation time */
                        if (xi+n1>=0 && xi+n2<nx && yi+n1>=0 && yi+n2<ny) {
                            adr = yi*nx+xi; 
                            for (dy=n1;dy<=n2;dy++) 
                                for (dx=n1;dx<=n2;dx++) 
                                /*		res += cy[n2-dy]*cx[n2-dx]*ref->gray[adr+nx*dy+dx];*/
                                    res += cy[n2-dy]*cx[n2-dx]*ref[adr+nx*dy+dx];
                        } else 
                            for (dy=n1;dy<=n2;dy++)
                                for (dx=n1;dx<=n2;dx++) 
                                /*		res += cy[n2-dy]*cx[n2-dx]*v(ref,xi+dx,yi+dy,*bg); */
                                    res += cy[n2-dy]*cx[n2-dx]*v(ref,xi+dx,yi+dy,*bg,nx,ny); 
                    }
                }		
                /* out->gray[y*(*sx)+x] = res; */
                out[y*(*sx)+x] = res;
            }
        //if (coeffs) 
        /* mw_delete_fimage(coeffs); */
        //  delete[] coeffs; 
    }

    
    
    
    
    
    /* --------------------------- Blur image --------------------------- */
    
    
    /* Same as ConvBuffer, but implemented with loop unrolling for increased
     speed.  This is the most time intensive routine in keypoint detection,
     so deserves careful attention to efficiency.  Loop unrolling simply
     sums 5 multiplications at a time to allow the compiler to schedule
     operations better and avoid loop overhead.  This almost triples
     speed of previous version on a Pentium with gcc.
     */
    void ConvBufferFast(float *buffer, float *kernel, int rsize, int ksize)
    {
        int i;
        float *bp, *kp, *endkp;
        float sum;
        
        for (i = 0; i < rsize; i++) {
            sum = 0.0;
            bp = &buffer[i];
            kp = &kernel[0];
            endkp = &kernel[ksize];
            
            /* Loop unrolling: do 5 multiplications at a time. */
            //      while (kp + 4 < endkp) {
            //      sum += (double) bp[0] * (double) kp[0] + (double)  bp[1] * (double) kp[1] + (double) bp[2] * (double) kp[2] +
            //             (double) bp[3] * (double) kp[3] + (double)  bp[4] * (double) kp[4];
            //      bp += 5;
            //      kp += 5;		  
            //     }
            //      /* Do 2 multiplications at a time on remaining items. */
            //     while (kp + 1 < endkp) {
            //       sum += (double) bp[0] * (double) kp[0] + (double)  bp[1] * (double) kp[1];
            //	   bp += 2;
            //	   kp += 2;
            //	 }
            //      /* Finish last one if needed. */
            //		if (kp < endkp) {
            //		sum += (double) *bp * (double) *kp;
            //		}
            
            while (kp < endkp) {
                sum += *bp++ * *kp++;			  
            }
            
            buffer[i] = sum;
        }
    }
    
    /* Convolve image with the 1-D kernel vector along image rows.  This
     is designed to be as efficient as possible.  Pixels outside the
     image are set to the value of the closest image pixel.
     */
    void ConvHorizontal(std::vector<float>& image, int width, int height, float *kernel, int ksize)
    {
        int rows, cols, r, c, i, halfsize;
        float buffer[4000];
	std::vector<float> pixels(width*height);
        
        
        rows = height;
        cols = width;
        
        halfsize = ksize / 2;
        pixels = image;
        assert(cols + ksize < 4000);
        
        for (r = 0; r < rows; r++) {
            /* Copy the row into buffer with pixels at ends replicated for
             half the mask size.  This avoids need to check for ends
             within inner loop. */
            for (i = 0; i < halfsize; i++)
                buffer[i] = pixels[r*cols];
            for (i = 0; i < cols; i++)
                buffer[halfsize + i] = pixels[r*cols+i];
            for (i = 0; i < halfsize; i++)
                buffer[halfsize + cols + i] = pixels[r*cols+cols-1];
            
            ConvBufferFast(buffer, kernel, cols, ksize);
            for (c = 0; c < cols; c++)
                pixels[r*cols+c] = buffer[c];
        }
        image = pixels;	 
    }
    
    
    /* Same as ConvHorizontal, but apply to vertical columns of image.
     */
    void ConvVertical(std::vector<float>& image, int width, int height, float *kernel, int ksize)
    {
        int rows, cols, r, c, i, halfsize;
        float buffer[4000];
	std::vector<float> pixels(width*height);
        
        rows = height;
        cols = width;
        
        halfsize = ksize / 2;
        pixels = image;
        assert(rows + ksize < 4000);
        
        for (c = 0; c < cols; c++) {
            for (i = 0; i < halfsize; i++)
                buffer[i] = pixels[c];
            for (i = 0; i < rows; i++)
                buffer[halfsize + i] = pixels[i*cols+c];
            for (i = 0; i < halfsize; i++)
                buffer[halfsize + rows + i] = pixels[(rows - 1)*cols+c];
            
            ConvBufferFast(buffer, kernel, rows, ksize);
            for (r = 0; r < rows; r++)
                pixels[r*cols+c] = buffer[r];
        }
        
        image = pixels;	 
    }
    
    
    
    /* 1D Convolve image with a Gaussian of width sigma and store result back
     in image.   This routine creates the Gaussian kernel, and then applies
     it in horizontal (flag_dir=0) OR vertical directions (flag_dir!=0).
     */
    void GaussianBlur1D(std::vector<float>& image, int width, int height, float sigma, int flag_dir)
    {
        float x, kernel[100], sum = 0.0;
        int ksize, i;
        
        /* The Gaussian kernel is truncated at GaussTruncate sigmas from
         center.  The kernel size should be odd.
         */
        ksize = (int)(2.0 * GaussTruncate1 * sigma + 1.0);
        ksize = MAX(3, ksize);    /* Kernel must be at least 3. */
        if (ksize % 2 == 0)       /* Make kernel size odd. */
            ksize++;
        assert(ksize < 100);
        
        /* Fill in kernel values. */
        for (i = 0; i <= ksize; i++) {
            x = i - ksize / 2;
            kernel[i] = exp(- x * x / (2.0 * sigma * sigma));
            sum += kernel[i];
        }
        /* Normalize kernel values to sum to 1.0. */
        for (i = 0; i < ksize; i++)
            kernel[i] /= sum;
        
        if (flag_dir == 0)
        {
            ConvHorizontal(image, width, height, kernel, ksize);
        }
        else
        {
            ConvVertical(image, width, height, kernel, ksize);
        }
    }
    
    
    void compensate_affine_coor1(float *x0, float *y0, int w1, int h1, float t1, float t2, float Rtheta)
    {
        float x_ori, y_ori;	
        float x_tmp, y_tmp;
        
        float x1 = *x0;
        float y1 = *y0;
        
        
        Rtheta = Rtheta*PI/180;
        
        if ( Rtheta <= PI/2 )
        {
            x_ori = 0;
            y_ori = w1 * sin(Rtheta) / t1;
        }
        else
        {
            x_ori = -w1 * cos(Rtheta) / t2;
            y_ori = ( w1 * sin(Rtheta) + h1 * sin(Rtheta-PI/2) ) / t1;
        }
        
        float sin_Rtheta = sin(Rtheta);
        float cos_Rtheta = cos(Rtheta);
        
        
        /* project the coordinates of im1 to original image before tilt-rotation transform */
        /* Get the coordinates with respect to the 'origin' of the original image before transform */
        x1 = x1 - x_ori;
        y1 = y1 - y_ori;
        /* Invert tilt */
        x1 = x1 * t2;
        y1 = y1 * t1;
        /* Invert rotation (Note that the y direction (vertical) is inverse to the usual concention. Hence Rtheta instead of -Rtheta to inverse the rotation.) */
        x_tmp = cos_Rtheta*x1 - sin_Rtheta*y1;
        y_tmp = sin_Rtheta*x1 + cos_Rtheta*y1;
        x1 = x_tmp;
        y1 = y_tmp;		
        
        *x0 = x1;
        *y0 = y1;
    }
    
    
    /* -------------- MAIN FUNCTION ---------------------- */
    
    int compute_asift_keypoints(std::vector<float>& image, int width, int height, int num_of_tilts, int verb, std::vector< std::vector< keypointslist > >& keys_all, siftPar &siftparameters)
    // Compute ASIFT keypoints in the input image.
    // Input:
    // image: input image
    // width, height: width and height of the input image.
    // num_of_tilts: number of tilts to simulate.
    // verb: 1/0 --> show/don not show verbose messages. (1 for debugging) 
    // keys_all (output): ASIFT keypoints. It is a 2D matrix with varying rows and columns. Each entry keys_all[tt][rr] 
    //	stores the SIFT keypoints calculated on the image with the simulated tilt index tt and simulated rotation index rr (see the code below). In the coordinates of the keypoints,
    //	the affine distortions have been compensated. 
    // siftparameters: SIFT parameters.
    //
    // Output: the number of keypoints 
    {	
	    std::vector<float> image_t, image_tmp1, image_tmp;	
        
        float t_min, t_k;
        int num_tilt, tt, num_rot_t2, rr;
        int fproj_o;
        float fproj_p, fproj_bg;
        char fproj_i;
        float *fproj_x4, *fproj_y4;
        //  float frot_b=0;
        float frot_b=128;
        char *frot_k;
        int  counter_sim=0, num_sim;
        int flag_dir = 1;
        float BorderFact=6*sqrt(2.);
        
        int num_keys_total=0;
        
        
        fproj_o = 3;
        fproj_p = 0;
        fproj_i = 0;
        fproj_bg = 0;
        fproj_x4 = 0;
        fproj_y4 = 0;
        
        frot_k = 0;
        
        num_rot_t2 = 10;
        
        t_min = 1;
        t_k = sqrt(2.);
        
        
        num_tilt = num_of_tilts;
        
        
        if ( num_tilt < 1)
        {
            printf("Number of tilts num_tilt should be equal or larger than 1. \n");
            exit(-1);	
        }
        
        image_tmp1 = image;
        
        
        /* Calculate the number of simulations, and initialize keys_all */
        keys_all = std::vector< std::vector< keypointslist > >(num_tilt);	
        for (tt = 1; tt <= num_tilt; tt++)
        {
            float t = t_min * pow(t_k, tt-1);
            
            if ( t == 1 )
            {
                counter_sim ++;
                
                keys_all[tt-1] = std::vector< keypointslist >(1);
            }
            else
            {
                int num_rot1 = round(num_rot_t2*t/2);        
                if ( num_rot1%2 == 1 )
                {
                    num_rot1 = num_rot1 + 1;
                }
                num_rot1 = num_rot1 / 2;
                counter_sim +=  num_rot1;
                
                keys_all[tt-1] = std::vector< keypointslist >(num_rot1);	
            }         		
        }
        
        num_sim = counter_sim;
        
        if ( verb )
        {
            printf("%d affine simulations will be performed. \n", num_sim);
        }
        
        counter_sim = 0;
        
        
        
        for (tt = 1; tt <= num_tilt; tt++)
        {
            float t = t_min * pow(t_k, tt-1);
            
            float t1 = 1;
            float t2 = 1/t;
            
            // If tilt t = 1, do not simulate rotation. 	
            if ( t == 1 )
            {					
                // copy the image from vector to array as compute_sift_keypoints uses only array.				
                float *image_tmp1_float = new float[width*height];
                for (int cc = 0; cc < width*height; cc++)
                    image_tmp1_float[cc] = image_tmp1[cc];
                
                compute_sift_keypoints(image_tmp1_float,keys_all[tt-1][0],width,height,siftparameters);
                
                delete[] image_tmp1_float;
                
            }
            else
            {
                // The number of rotations to simulate under the current tilt.   
                int num_rot1 = round(num_rot_t2*t/2);        
                
                if ( num_rot1%2 == 1 )
                {
                    num_rot1 = num_rot1 + 1;
                }
                num_rot1 = num_rot1 / 2;
                float delta_theta = PI/num_rot1;		
                
     
                
                for ( int rr = 1; rr <= num_rot1; rr++ ) 
                {
                    float theta = delta_theta * (rr-1);
                    theta = theta * 180 / PI;
                    
		    std::vector<float> image_t;
                    int width_r, height_r;
                    
                    // simulate a rotation: rotate the image with an angle theta. (the outside of the rotated image are padded with the value frot_b)
                    frot(image, image_t, width, height, &width_r, &height_r, &theta, &frot_b , frot_k);
                    
                    /* Tilt */			 
                    int width_t = (int) (width_r * t1);
                    int height_t = (int) (height_r * t2);  
                    
                    int fproj_sx = width_t;
                    int fproj_sy = height_t;     
                    
                    float fproj_x1 = 0;
                    float fproj_y1 = 0;
                    float fproj_x2 = width_t;
                    float fproj_y2 = 0;
                    float fproj_x3 = 0;	     
                    float fproj_y3 = height_t;
                    
                    /* Anti-aliasing filtering along vertical direction */
                    /* sigma_aa = InitSigma_aa * log2(t);*/
                    float sigma_aa = InitSigma_aa * t / 2;
                    GaussianBlur1D(image_t,width_r,height_r,sigma_aa,flag_dir);
                    
                    
                    // simulate a tilt: subsample the image along the vertical axis by a factor of t.
		    std::vector<float> image_tmp(width_t*height_t);			 
                    fproj (image_t, image_tmp, width_r, height_r, &fproj_sx, &fproj_sy, &fproj_bg, &fproj_o, &fproj_p, &fproj_i , fproj_x1 , fproj_y1 , fproj_x2 , fproj_y2 , fproj_x3 , fproj_y3, fproj_x4, fproj_y4); 
                    
		    std::vector<float> image_tmp1 = image_tmp;	 
                    
                    if ( verb )
                    {
                        printf("Rotation theta = %.2f, Tilt t = %.2f. w=%d, h=%d, sigma_aa=%.2f, \n", theta, t, width_t, height_t, sigma_aa);
                    }
                    
                    
                    float *image_tmp1_float = new float[width_t*height_t];
                    for (int cc = 0; cc < width_t*height_t; cc++)
                        image_tmp1_float[cc] = image_tmp1[cc];	 
                    
                    // compute SIFT keypoints on simulated image. 	 
                    keypointslist keypoints;
                    keypointslist keypoints_filtered;
                    compute_sift_keypoints(image_tmp1_float,keypoints,width_t,height_t,siftparameters);
                    
                    delete[] image_tmp1_float;		
                    
                    /* check if the keypoint is located on the boundary of the parallelogram (i.e., the boundary of the distorted input image). If so, remove it to avoid boundary artifacts. */
                    if ( keypoints.size() != 0 )
                    {
                        for ( int cc = 0; cc < (int) keypoints.size(); cc++ )
                        {		      
                            
                            float x0, y0, x1, y1, x2, y2, x3, y3 ,x4, y4, d1, d2, d3, d4, scale1, theta1, sin_theta1, cos_theta1, BorderTh;
                            
                            x0 = keypoints[cc].x;
                            y0 = keypoints[cc].y;
                            scale1= keypoints[cc].scale;
                            
                            theta1 = theta * PI / 180;
                            sin_theta1 = sin(theta1);
                            cos_theta1 = cos(theta1);
                            
                            /* the coordinates of the 4 submits of the parallelogram */
                            if ( theta <= 90 )
                            {
                                x1 = height * sin_theta1;
                                y1 = 0;			 
                                y2 = width * sin_theta1;
                                x3 = width * cos_theta1;
                                x4 = 0;
                                y4 = height * cos_theta1;
                                x2 = x1 + x3;
                                y3 = y2 + y4;
                                
                                /* note that the vertical direction goes from top to bottom!!! 
                                 The calculation above assumes that the vertical direction goes from the bottom to top. Thus the vertical coordinates need to be reversed!!! */
                                y1 = y3 - y1;
                                y2 = y3 - y2;
                                y4 = y3 - y4;
                                y3 = 0;
                                
                                y1 = y1 * t2;
                                y2 = y2 * t2;
                                y3 = y3 * t2;
                                y4 = y4 * t2;
                            }
                            else
                            {
                                y1 = -height * cos_theta1;
                                x2 = height * sin_theta1;
                                x3 = 0;
                                y3 = width * sin_theta1;				 
                                x4 = -width * cos_theta1;
                                y4 = 0;
                                x1 = x2 + x4;
                                y2 = y1 + y3;
                                
                                /* note that the vertical direction goes from top to bottom!!! 
                                 The calculation above assumes that the vertical direction goes from the bottom to top. Thus the vertical coordinates need to be reversed!!! */
                                y1 = y2 - y1;
                                y3 = y2 - y3;
                                y4 = y2 - y4;
                                y2 = 0;
                                
                                y1 = y1 * t2;
                                y2 = y2 * t2;
                                y3 = y3 * t2;
                                y4 = y4 * t2;
                            }		       		    
                            
                            /* the distances from the keypoint to the 4 sides of the parallelogram */
                            d1 = ABS((x2-x1)*(y1-y0)-(x1-x0)*(y2-y1)) / sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1));
                            d2 = ABS((x3-x2)*(y2-y0)-(x2-x0)*(y3-y2)) / sqrt((x3-x2)*(x3-x2)+(y3-y2)*(y3-y2));
                            d3 = ABS((x4-x3)*(y3-y0)-(x3-x0)*(y4-y3)) / sqrt((x4-x3)*(x4-x3)+(y4-y3)*(y4-y3));
                            d4 = ABS((x1-x4)*(y4-y0)-(x4-x0)*(y1-y4)) / sqrt((x1-x4)*(x1-x4)+(y1-y4)*(y1-y4));
                            
                            BorderTh = BorderFact*scale1;
                            
                            if (!((d1<BorderTh) || (d2<BorderTh) || (d3<BorderTh) || (d4<BorderTh) ))
                            {				 					   
                                // Normalize the coordinates of the matched points by compensate the simulate affine transformations
                                compensate_affine_coor1(&x0, &y0, width, height, 1/t2, t1, theta);
                                keypoints[cc].x = x0;
                                keypoints[cc].y = y0;
                                
                                keypoints_filtered.push_back(keypoints[cc]);	 
                            }				   
                        }
                    }			 
                    keys_all[tt-1][rr-1] = keypoints_filtered;
                }		 
            }         
        }
        
        {		
            for (tt = 0; tt < (int) keys_all.size(); tt++)
                for (rr = 0; rr < (int) keys_all[tt].size(); rr++)
                {
                    num_keys_total += (int) keys_all[tt][rr].size();
                }				
            printf("%d ASIFT keypoints are detected. \n", num_keys_total);
        }
        
        return num_keys_total;
    }

    
    
    
    
    
    /* Remove the repetitive matches that appear in different simulations and retain only one */
    void unique_match1(matchingslist &seg_in, matchingslist &seg_out, std::vector< std::vector <float> > &Minfoall_in, std::vector< std::vector <float> > &Minfoall_out)
    {
        int i_in, i_out;
        float x1_in, x2_in, y1_in, y2_in, x1_out, x2_out, y1_out, y2_out;
        int flag_unique;
        float d1, d2;
        int Th2 = 2;
        
        seg_out.push_back(seg_in[0]);
        Minfoall_out.push_back(Minfoall_in[0]);	
        
        /* For other matches */
        if ( seg_in.size() > 1 )	
        {
            /* check if a match is unique. if yes, copy */
            matchingslist::iterator ptr_in = seg_in.begin();
            for ( i_in = 1; i_in < (int) seg_in.size(); i_in++, ptr_in++ )
			{				
                x1_in = ptr_in->first.x;
                y1_in = ptr_in->first.y;
                x2_in = ptr_in->second.x; 
                y2_in = ptr_in->second.y;	
                
                flag_unique = 1;
                
                matchingslist::iterator ptr_out = seg_out.begin(); 
                for ( i_out = 0; i_out < (int) seg_out.size(); i_out++, ptr_out++ )
				{
					x1_out = ptr_out->first.x;
					y1_out = ptr_out->first.y;
					x2_out = ptr_out->second.x; 
					y2_out = ptr_out->second.y;	
					
					d1 = (x1_in - x1_out)*(x1_in - x1_out) + (y1_in - y1_out)*(y1_in - y1_out);
					d2 = (x2_in - x2_out)*(x2_in - x2_out) + (y2_in - y2_out)*(y2_in - y2_out);
                    
                    
					if ( ( d1 <= Th2) && ( d2 <= Th2) )
                    {
                        flag_unique = 0;
                        continue;
                    }	       
				}
                
                if ( flag_unique == 1 )
				{									
					seg_out.push_back(seg_in[i_in]);
					Minfoall_out.push_back(Minfoall_in[i_in]);						
				}	  
            }
        }
    }
    
    /* Remove the ALL one-to-multiple matches. */
    void clean_match1(matchingslist &seg_in, matchingslist &seg_out, std::vector< std::vector <float> > &Minfoall_in, std::vector< std::vector <float> > &Minfoall_out)
    {
        int i1, i2;
        float x1_in, x2_in, y1_in, y2_in, x1_out, x2_out, y1_out, y2_out;
        
        // Guoshen Yu, 2010.09.22, Windows version
        // int flag_unique[seg_in.size()];
        int tmp_size = seg_in.size();
        int *flag_unique = new int[tmp_size];
        
        int sum_flag=0;
        float d1, d2;
        int Th1 = 1;
        int Th2 = 4;
        
        for ( i1 = 0; i1 < (int) seg_in.size(); i1++ )
        {
            flag_unique[i1] = 1;
        }
        
        /* Set the flag of redundant matches to 0. */
        matchingslist::iterator ptr_in = seg_in.begin();
        for ( i1 = 0; i1 < (int) seg_in.size() - 1; i1++, ptr_in++ )
        {					
            x1_in = ptr_in->first.x;
            y1_in = ptr_in->first.y;
            x2_in = ptr_in->second.x; 
            y2_in = ptr_in->second.y;
            
            matchingslist::iterator ptr_out = ptr_in+1;
            for ( i2 = i1 + 1; i2 < (int) seg_in.size(); i2++, ptr_out++ )
            {
                x1_out = ptr_out->first.x;
                y1_out = ptr_out->first.y;
                x2_out = ptr_out->second.x; 
                y2_out = ptr_out->second.y;	
                
                d1 = (x1_in - x1_out)*(x1_in - x1_out) + (y1_in - y1_out)*(y1_in - y1_out);
                d2 = (x2_in - x2_out)*(x2_in - x2_out) + (y2_in - y2_out)*(y2_in - y2_out);
                
                /* If redundant, set flags of both elements to 0.*/
                if ( ( ( d1 <= Th1) && ( d2 > Th2) ) || ( ( d1 > Th2) && ( d2 <= Th1) ) )
                {
                    flag_unique[i1] = 0;
                    flag_unique[i2] = 0;				  
                }
            }		
        }
        
        for ( i1 = 0; i1 < (int) seg_in.size(); i1++ )
        {
            sum_flag += flag_unique[i1];
        }
        
        /* Copy the matches that are not redundant */
        if ( sum_flag > 0 )
        {
            for (  i1 = 0; i1 < (int) seg_in.size(); i1++ )
            {
                if ( flag_unique[i1] == 1 )
                {				
                    seg_out.push_back(seg_in[i1]);
                    Minfoall_out.push_back(Minfoall_in[i1]);   
                }
            }
        }
        else
        {
            printf("Warning: all matches are redundant and are thus removed! This step of match cleaning is short circuited. (Normally this should not happen...)\n");
        }
        
        // Guoshen Yu, 2010.09.22, Windows version
        delete [] flag_unique;
    }
    
    
    /* Remove the ALL multiple-to-one matches */
    void clean_match2(matchingslist &seg_in, matchingslist &seg_out, std::vector< std::vector <float> > &Minfoall_in, std::vector< std::vector <float> > &Minfoall_out)
    {
        int i1, i2;
        float x1_in, x2_in, y1_in, y2_in, x1_out, x2_out, y1_out, y2_out;
        
        // Guoshen Yu, 2010.09.22, Windows version
        // int flag_unique[seg_in.size()];
        int tmp_size = seg_in.size();
        int *flag_unique = new int[tmp_size];
        
        int sum_flag=0;
        float d1, d2;
        int Th1 = 1;
        int Th2 = 4;
        
        for ( i1 = 0; i1 < (int) seg_in.size(); i1++ )
        {
            flag_unique[i1] = 1;
        }
        
        /* Set the flag of redundant matches to 0. */
        matchingslist::iterator ptr_in = seg_in.begin();
        for ( i1 = 0; i1 < (int) seg_in.size() - 1; i1++, ptr_in++ )
        {																				
            x1_in = ptr_in->first.x;
            y1_in = ptr_in->first.y;
            x2_in = ptr_in->second.x; 
            y2_in = ptr_in->second.y;	
            
            matchingslist::iterator ptr_out = ptr_in+1;
            for ( i2 = i1 + 1; i2 < (int) seg_in.size(); i2++, ptr_out++ )
            {
                
                x1_out = ptr_out->first.x;
                y1_out = ptr_out->first.y;
                x2_out = ptr_out->second.x; 
                y2_out = ptr_out->second.y;	
                
                d1 = (x1_in - x1_out)*(x1_in - x1_out) + (y1_in - y1_out)*(y1_in - y1_out);
                d2 = (x2_in - x2_out)*(x2_in - x2_out) + (y2_in - y2_out)*(y2_in - y2_out);
                
                
                /* If redundant, set flags of both elements to 0.*/	
                if ( ( d1 > Th2) && ( d2 <= Th1) ) 
                {
                    flag_unique[i1] = 0;
                    flag_unique[i2] = 0;				  
                }
            }
        }
        
        for ( i1 = 0; i1 < (int) seg_in.size(); i1++ )
        {
            sum_flag += flag_unique[i1];
        }
        
        /* Copy the matches that are not redundant */
        if ( sum_flag > 0 )
        {
            for (  i1 = 0; i1 < (int) seg_in.size(); i1++ )
            {
                if ( flag_unique[i1] == 1 )
                {			 
                    seg_out.push_back(seg_in[i1]);
                    Minfoall_out.push_back(Minfoall_in[i1]);    
                }
            }
        }
        else
        {
            printf("Warning: all matches are redundant and are thus removed! This step of match cleaning is short circuited. (Normally this should not happen...)\n");
        }
        
        // Guoshen Yu, 2010.09.22, Windows version
        delete [] flag_unique;
    }
    
    
    // Normalize the coordinates of the matched points by compensating the simulate affine transformations
    void compensate_affine_coor(matching &matching1, int w1, int h1, int w2, int h2, float t1, float t2, float Rtheta, float t_im2_1, float t_im2_2, float Rtheta2)
    {
        float x_ori, y_ori;	
        float x_ori2, y_ori2, x_tmp, y_tmp;
        float x1, y1, x2, y2;
		
        Rtheta = Rtheta*PI/180;
        
        if ( Rtheta <= PI/2 )
        {
            x_ori = 0;
            y_ori = w1 * sin(Rtheta) / t1;
        }
        else
        {
            x_ori = -w1 * cos(Rtheta) / t2;
            y_ori = ( w1 * sin(Rtheta) + h1 * sin(Rtheta-PI/2) ) / t1;
        }
        
        Rtheta2 = Rtheta2*PI/180;
        
        if ( Rtheta2 <= PI/2 )
        {
            x_ori2 = 0;
            y_ori2 = w2 * sin(Rtheta2) / t_im2_1;
        }
        else
        {
            x_ori2 = -w2 * cos(Rtheta2) / t_im2_2;
            y_ori2 = ( w2 * sin(Rtheta2) + h2 * sin(Rtheta2-PI/2) ) / t_im2_1;
        }
        
        float sin_Rtheta = sin(Rtheta);
        float cos_Rtheta = cos(Rtheta);
        float sin_Rtheta2 = sin(Rtheta2);
        float cos_Rtheta2 = cos(Rtheta2);
        
        x1 = matching1.first.x;
        y1 = matching1.first.y;	
        x2 = matching1.second.x;
        y2 = matching1.second.y;
        
        /* project the coordinates of im1 to original image before tilt-rotation transform */
        /* Get the coordinates with respect to the 'origin' of the original image before transform */
        x1 = x1 - x_ori;
        y1 = y1 - y_ori;
        /* Invert tilt */
        x1 = x1 * t2;
        y1 = y1 * t1;
        /* Invert rotation (Note that the y direction (vertical) is inverse to the usual concention. Hence Rtheta instead of -Rtheta to inverse the rotation.) */
        x_tmp = cos_Rtheta*x1 - sin_Rtheta*y1;
        y_tmp = sin_Rtheta*x1 + cos_Rtheta*y1;
        x1 = x_tmp;
        y1 = y_tmp;
        
        /* Coordinate projection on image2  */
        
        /* Get the coordinates with respect to the 'origin' of the original image before transform */
        x2 = x2 - x_ori2;
        y2 = y2 - y_ori2;
        /* Invert tilt */
        x2 = x2 * t_im2_2;
        y2 = y2 * t_im2_1;
        /* Invert rotation (Note that the y direction (vertical) is inverse to the usual concention. Hence Rtheta instead of -Rtheta to inverse the rotation.) */
        x_tmp = cos_Rtheta2*x2 - sin_Rtheta2*y2;
        y_tmp = sin_Rtheta2*x2 + cos_Rtheta2*y2;
        x2 = x_tmp;
        y2 = y_tmp;
        
        matching1.first.x = x1;
        matching1.first.y = y1;	
        matching1.second.x = x2;
        matching1.second.y = y2;
    }
    
    int compute_asift_matches(int num_of_tilts1, int num_of_tilts2,  int verb, std::vector< std::vector< keypointslist > >& keys1, std::vector< std::vector< keypointslist > >& keys2, matchingslist &matchings, siftPar &siftparameters)
    // Match the ASIFT keypoints. 
    // Input:
    // num_of_tilts1, num_of_tilts2: number of tilts that have been simulated on the two images. (They can be different.)
    // w1, h1, w2, h2: widht/height of image1/image2.
    // verb: 1/0 --> show/don not show verbose messages. (1 for debugging) 
    // keys1, keys2: ASIFT keypoints of image1/image2. (They should be calculated with compute_asift_keypoints.)
    // matchings (output): the coordinates (col1, row1, col2, row2) of all the matching points. 
    //
    // Output: the number of matching points.
    {  	
        float t_min, t_k, t;
        int num_tilt1, num_tilt2, tt, num_rot_t2, num_rot1, rr;
        int cc;
        
        int tt2, rr2, num_rot1_2;
        float t_im2;
        
        /* It stores the coordinates of ALL matches points of ALL affine simulations  */
	std::vector< std::vector <float> > Minfoall;
        
        int Tmin = 8;					
        
        num_rot_t2 = 10;
        
        t_min = 1;
        t_k = sqrt(2.);
        
        num_tilt1 = num_of_tilts1;
        num_tilt2 = num_of_tilts2;
        
        if ( ( num_tilt1 < 1 ) || ( num_tilt2 < 1 ) )
        {
            printf("Number of tilts num_tilt should be equal or larger than 1. \n");
            exit(-1);	
        }
        
        
        /* Initialize the vector structure for the matching points */	
        std::vector< std::vector< std::vector < std::vector < matchingslist > > > > matchings_vec(num_tilt1);	
        std::vector< std::vector< std::vector< std::vector< std::vector< std::vector <float> > > > > > Minfoall_vec(num_tilt1);	
        for (tt = 1; tt <= num_tilt1; tt++)
        {		
            t = t_min * pow(t_k, tt-1);
            if ( t == 1 )
            {			
                num_rot1 = 1;
            }
            else
            {
                num_rot1 = round(num_rot_t2*t/2);        
                if ( num_rot1%2 == 1 )
                {
                    num_rot1 = num_rot1 + 1;
                }
                num_rot1 = num_rot1 / 2;
            }     
            
            matchings_vec[tt-1].resize(num_rot1);
            Minfoall_vec[tt-1].resize(num_rot1);
            
            for  ( rr = 1; rr <= num_rot1; rr++ ) 
            {   
                
                matchings_vec[tt-1][rr-1].resize(num_tilt2);
                Minfoall_vec[tt-1][rr-1].resize(num_tilt2);
                
                for (tt2 = 1; tt2 <= num_tilt2; tt2++)				
                {			
                    t_im2 = t_min * pow(t_k, tt2-1);
                    if ( t_im2 == 1 )
                    {
                        num_rot1_2 = 1;
                    }
                    else
                    {
                        num_rot1_2 = round(num_rot_t2*t_im2/2);        
                        if ( num_rot1_2%2 == 1 )
                        {
                            num_rot1_2 = num_rot1_2 + 1;
                        }
                        num_rot1_2 = num_rot1_2 / 2;
                    }
                    
                    matchings_vec[tt-1][rr-1][tt2-1].resize(num_rot1_2);		
                    Minfoall_vec[tt-1][rr-1][tt2-1].resize(num_rot1_2);				
                }
            }
        }
        
        
        ///*
        // * setup the tilt and rotation parameters
        // * for all the loops, this vector will hold
        // * the following parameters: 
        // * tt, num_rot1, rr, tt2, num_rot1_2, rr2
        // */
        //vector<int> tilt_rot;
        ///* loop on tilts for image 1 */ 
        //for (int tt = 1; tt <= num_tilt1; tt++)
        //{
        //    float t = t_min * pow(t_k, tt-1);
        //    int num_rot1;
        //    /* if tilt t = 1, do not simulate rotation. */
        //    if ( 1 == tt )
        //		num_rot1 = 1;
        //    else
        //    {
        //		/* number of rotations to simulate */
        //		num_rot1 = round(num_rot_t2 * t / 2);        
        //		if ( num_rot1%2 == 1 )
        //			num_rot1 = num_rot1 + 1;
        //		num_rot1 = num_rot1 / 2;
        //    }
        //    /* loop on rotations for image 1 */ 
        //    for  (int rr = 1; rr <= num_rot1; rr++ ) 
        //    {   
        //		/* loop on tilts for image 2 */
        //		for (int tt2 = 1; tt2 <= num_tilt2; tt2++)
        //		{
        //			float t_im2 = t_min * pow(t_k, tt2-1);
        //			int num_rot1_2;
        //			if ( tt2 == 1 )
        //				num_rot1_2 = 1;
        //			else
        //			{
        //				num_rot1_2 = round(num_rot_t2 * t_im2 / 2);        
        //				if ( num_rot1_2%2 == 1 )
        //					num_rot1_2 = num_rot1_2 + 1;
        //				num_rot1_2 = num_rot1_2 / 2;
        //			}
        //			/* loop on rotations for image 2 */ 
        //			for  (int rr2 = 1; rr2 <= num_rot1_2; rr2++ ) 
        //			{
        //				tilt_rot.push_back(tt);
        //				tilt_rot.push_back(num_rot1);
        //				tilt_rot.push_back(rr);
        //				tilt_rot.push_back(tt2);
        //				tilt_rot.push_back(num_rot1_2);
        //				tilt_rot.push_back(rr2);
        //			}
        //		}
        //    }
        //}
		

        
        for (int tt = 1; tt <= num_tilt1; tt++)
        {
            
            float t = t_min * pow(t_k, tt-1);
            
            /* Attention: the t1, t2 do not follow the same convention as in compute_asift_keypoints */   
            float t1 = t; 
            float t2 = 1;
            
            int num_rot1;
            
            // If tilt t = 1, do not simulate rotation. 	
            if ( tt == 1 )
            {
                num_rot1 = 1;
            }
            else
            {
                // The number of rotations to simulate under the current tilt.   
                num_rot1 = round(num_rot_t2*t/2);        
                if ( num_rot1%2 == 1 )
                {
                    num_rot1 = num_rot1 + 1;
                }
                num_rot1 = num_rot1 / 2;
            }
            
            
            float delta_theta = PI/num_rot1;
            

            
            for  ( int rr = 1; rr <= num_rot1; rr++ ) 
            {   
                float theta = delta_theta * (rr-1);
                theta = theta * 180 / PI;
                
                /* Read the keypoints of image 1 */	       
                keypointslist keypoints1 = keys1[tt-1][rr-1];
                
                // loop on tilts for image 2.

                for (int tt2 = 1; tt2 <= num_tilt2; tt2++)				
                {
                    float t_im2 = t_min * pow(t_k, tt2-1);
                    
                    /* Attention: the t1, t2 do not follow the same convention as in asift_v1.c */
                    float t_im2_1 = t_im2; 
                    float t_im2_2 = 1;
                    
                    int num_rot1_2;
                    
                    if ( tt2 == 1 )
                    {
                        num_rot1_2 = 1;
                    }
                    else
                    {
                        num_rot1_2 = round(num_rot_t2*t_im2/2);        
                        if ( num_rot1_2%2 == 1 )
                        {
                            num_rot1_2 = num_rot1_2 + 1;
                        }
                        num_rot1_2 = num_rot1_2 / 2;
                    }
                    
                    float delta_theta2 = PI/num_rot1_2;

                    
                    // Loop on rotations for image 2. 
                    for  ( int rr2 = 1; rr2 <= num_rot1_2; rr2++ ) 
                    {   
                        float theta2 = delta_theta2 * (rr2-1);
                        theta2 = theta2 * 180 / PI;
                        
                        /* Read the keypoints of image2. */	       
                        keypointslist keypoints2 = keys2[tt2-1][rr2-1];
                        
                        
                        // Match the keypoints of image1 and image2.
                        matchingslist matchings1;					
                        compute_sift_matches(keypoints1,keypoints2,matchings1,siftparameters);		       
                        
                        if ( verb ) 
                        {
                            printf("t1=%.2f, theta1=%.2f, num keys1 = %d, t2=%.2f, theta2=%.2f, num keys2 = %d, num matches=%d\n", t, theta, (int) keypoints1.size(), t_im2, theta2, (int) keypoints2.size(), (int) matchings1.size());
                        }
                        
                        /* Store the matches */
                        if ( matchings1.size() > 0 )
                        {
                            matchings_vec[tt-1][rr-1][tt2-1][rr2-1] = matchingslist(matchings1.size());
                            Minfoall_vec[tt-1][rr-1][tt2-1][rr2-1].resize(matchings1.size());
                            
                            for ( int cc = 0; cc < (int) matchings1.size(); cc++ )		   
                            {		      														
                                ///// In the coordinates the affine transformations have been normalized already in compute_asift_keypoints. So no need to normalize here. 
                                // Normalize the coordinates of the matched points by compensating the simulate affine transformations
                                //	compensate_affine_coor(matchings1[cc], w1, h1, w2, h2, t1, t2, theta, t_im2_1, t_im2_2, theta2);
                                
                                matchings_vec[tt-1][rr-1][tt2-1][rr2-1][cc] = matchings1[cc];
                                
				std::vector<float> Minfo_1match(6);
                                Minfo_1match[0] = t1;
                                Minfo_1match[1] = t2;
                                Minfo_1match[2] = theta;
                                Minfo_1match[3] = t_im2_1;
                                Minfo_1match[4] = t_im2_2;
                                Minfo_1match[5] = theta2;									
                                Minfoall_vec[tt-1][rr-1][tt2-1][rr2-1][cc] = Minfo_1match;
                            } 												
                        }
                    }
                }
            }
        }
        
        // Move the matches to a 1D vector
        for (tt = 1; tt <= num_tilt1; tt++)
        {				
            t = t_min * pow(t_k, tt-1);
            
            if ( t == 1 )
            {			
                num_rot1 = 1;
            }
            else
            {
                num_rot1 = round(num_rot_t2*t/2);        
                if ( num_rot1%2 == 1 )
                {
                    num_rot1 = num_rot1 + 1;
                }
                num_rot1 = num_rot1 / 2;
            }     
            
            for  ( rr = 1; rr <= num_rot1; rr++ ) 
            {   			
                for (tt2 = 1; tt2 <= num_tilt2; tt2++)				
                {			
                    t_im2 = t_min * pow(t_k, tt2-1);
                    if ( t_im2 == 1 )
                    {
                        num_rot1_2 = 1;
                    }
                    else
                    {
                        num_rot1_2 = round(num_rot_t2*t_im2/2);        
                        if ( num_rot1_2%2 == 1 )
                        {
                            num_rot1_2 = num_rot1_2 + 1;
                        }
                        num_rot1_2 = num_rot1_2 / 2;
                    }
                    
                    for  ( rr2 = 1; rr2 <= num_rot1_2; rr2++ ) 
                    { 
                        for ( cc=0; cc < (int) matchings_vec[tt-1][rr-1][tt2-1][rr2-1].size(); cc++ )
                        {
                            matchings.push_back(matchings_vec[tt-1][rr-1][tt2-1][rr2-1][cc]);
                            Minfoall.push_back(Minfoall_vec[tt-1][rr-1][tt2-1][rr2-1][cc]);
                        }
                    }
                }
            }
        }
        
        if ( verb )
        {
            printf("The number of matches is %d \n", (int) matchings.size());
        }
        
        
        if ( matchings.size() > 0 )
        {
            /* Remove the repetitive matches that appear in different simulations and retain only one. */		
            // Since tilts are simuated on both image 1 and image 2, it is normal to have repetitive matches. 
            matchingslist matchings_unique;	
	    std::vector< std::vector<float> > Minfoall_unique;			
            unique_match1(matchings, matchings_unique, Minfoall, Minfoall_unique);      		
            matchings = matchings_unique;
            Minfoall = Minfoall_unique; 
            
            if ( verb )    
			{
                printf("The number of unique matches is %d \n", (int) matchings.size());
			}
            
            // There often appear to be some one-to-multiple/multiple-to-one matches (one point in image 1 matches with many points in image 2/vice versa). 
            // This is an artifact of SIFT on interpolated images, as the interpolation tends to create some auto-similar structures (steps for example). 
            // These matches need to be removed.		
            /* Separating the removal of  multiple-to-one and one-to-multiple in two steps:
             - first remove multiple-to-one
             - then remove one-to-multiple
             This allows to avoid removing some good matches: multiple-to-one matches is much more frequent than one-to-multiple. Sometimes some of the feature points in image 1 that take part in "multiple-to-one" bad matches have also correct matches in image 2. The modified scheme avoid removing these good matches. */
            
            // Remove to multiple-to-one matches	
            matchings_unique.clear();
            Minfoall_unique.clear();		
            clean_match2(matchings, matchings_unique, Minfoall, Minfoall_unique);    		
            matchings = matchings_unique;
            Minfoall = Minfoall_unique;
            
            // Remove to one-to-multiple matches	
            matchings_unique.clear();
            Minfoall_unique.clear();		
            clean_match1(matchings, matchings_unique, Minfoall, Minfoall_unique);    		
            matchings = matchings_unique;
            Minfoall = Minfoall_unique;
            
            
            if ( verb )
            {
                printf("The number of final matches is %d \n", (int) matchings.size());
            }
            
            
            
            
            // If enough matches to do epipolar filtering
            /*if ( (int) matchings.size() >= Tmin )
            {			
                //////// Use ORSA to filter out the incorrect matches. 
                // store the coordinates of the matching points
                vector<Match> match_coor; 
                for ( cc = 0; cc < (int) matchings.size(); cc++ )	
                {
                    Match match1_coor;
                    match1_coor.x1 = matchings[cc].first.x;
                    match1_coor.y1 = matchings[cc].first.y;
                    match1_coor.x2 = matchings[cc].second.x;
                    match1_coor.y2 = matchings[cc].second.y;
                    
                    match_coor.push_back(match1_coor);							
                }
                
                std::vector<float> index;
                // Guoshen Yu, 2010.09.23
                // index.clear();
                
        
                
                
                
          
                    // extract meaningful matches
                    matchings_unique.clear();
                    Minfoall_unique.clear();
                    for ( cc = 0; cc < (int) index.size(); cc++ )				
                    {
                        matchings_unique.push_back(matchings[(int)index[cc]]);
                        Minfoall_unique.push_back(Minfoall[(int)index[cc]]);
                    }
                    matchings = matchings_unique;
                    Minfoall = Minfoall_unique;     
                    
                
            }
            else 
            {
                matchings.clear();
                Minfoall.clear();
                cout << "The two images do not match. Not enough matches to do epipolar filtering." << endl;		
            }
            */
        }
        else 
        {
		std::cout << "The two images do not match.\n" << std::endl;
        }
        
        return matchings.size();
        
    }
    

    
    
    
}
